"""
This file implements YAML calibration, a mechanism of programming multiple automated calibration steps
(and related operations). Essentially this scheme coordinates repeated calls to at.calibrate with different
adjustables and measurables in a pre-defined sequence of automated calibration steps.
"""

import sciris as sc
from pathlib import Path
import shutil
import atomica as at
import numpy as np
import yaml
import time
import re

__all__ = ["build", "run"]

from atomica import ParameterSet


def _get_named_nodes():
    """
    Return dictionary with all named Node subclasses
    """
    return {x._name: x for x in BaseNode.__subclasses__() if x._name is not None}


def build(instructions=None, context=None, name="calibration"):
    """
    Construct nodes representing a calibration

    :param instructions: A dictionary of attributes/settings defined for this node OR a string filename
                         containing a YAML file that can be loaded to provide instructions
    :param context: A dictionary of attributes/settings inherited from parent nodes
    :param name: The name to assign this node
    :param fname: Optionally read the instructions from a file
    :return: A Node subclass instance, the type of which depends on the instructions
    """

    if (sc.isstring(instructions) or isinstance(instructions, Path)) and Path(instructions).exists():
        with open(instructions) as file:
            instructions = yaml.load(file, Loader=yaml.FullLoader)

    named_nodes = _get_named_nodes()
    if isinstance(instructions, dict) and ("adjustables" in instructions or (context is not None and "adjustables" in context)) and ("measurables" in instructions or (context is not None and "measurables" in context)):
        return CalibrationNode(instructions, context, name)
    elif name in named_nodes:
        return named_nodes[name](instructions, context, name)
    else:
        return SectionNode(instructions, context, name)


def run(node, project, parset, savedir=None, save_intermediate=False, log_output: bool = False, *args, **kwargs):
    """
    Run YAML calibration

    This will execute the YAML calibration using the passed-in node (or instructions to build a node), and any associated children

    :param node: Calibration node to execute. If not a node (i.e., a YAML file, or node instructions), it will be converted into a node
    :param P: Project to which to apply these instructions
    :param parset: An `at.ParameterSet` instance to calibrate
    :param savedir: Optionally specify a directory to save the results. Defaults to the current working directory
    :param save_intermediate: Set whether to save intermediate calibrations (defaults to False)
    :return new_parset: A calibrated `at.ParameterSet` instance
    """

    if savedir is None:
        savedir = Path(".")
    else:
        savedir = Path(savedir)
    savedir.mkdir(exist_ok=True, parents=True)

    if not isinstance(node, BaseNode):
        # Save a copy of the yaml-file if saving log output
        if isinstance(node, Path) and log_output:
            shutil.copyfile(node, savedir / node.name)
        node = build(node)

    parset = sc.dcp(project.parset(parset))

    nodes = list(node.walk())  # Make a flat list of all nodes to execute in order
    n_steps = len([x for x in nodes if not isinstance(x[1], SectionNode)])
    n = 1

    if log_output:
        at.start_logging(savedir / "calibration_log.txt")

    at.logger.info(f"\nStarting calibration ({n_steps} steps)")

    for n_reps, node in nodes:

        if isinstance(node, SectionNode):
            at.logger.info(f'\nSection "{node.name}" (repeat {n_reps} of {node.repeats})')
        else:
            at.logger.info(f'\nStep {n} of {n_steps}: "{node.name}" (repeat {n_reps} of {node.repeats})')
            parset = node.apply(project, parset, savedir, save_intermediate, *args, **kwargs)
            n += 1

            if save_intermediate and not isinstance(node, SaveCalibrationNode):
                output = savedir / f'intermediate_calibration_{n:0{len(str(n_steps))}}_{node.name.replace(" ", "_")}'
                at.logger.info(f"Saving intermediate calibration...")
                parset.save_calibration(output)

    t = time.process_time()
    at.logger.info(f"\nCalibration completed. Total time elapsed: {round(t, 2)} seconds ({round(t/60, 2)} minutes)")

    if log_output:
        at.stop_logging()

    return parset


class BaseNode:
    """
    Node base class

    The base node class implements basic node features. Typically there should not be any
    instances of this class, only instances of subclasses
    """

    _name = None  # If specified, this key can be used as the name of the step to create a node of this type

    def __init__(self, instructions, context, name):
        self.name = name
        self.instructions = sc.dcp(instructions)
        self.context = context  # Attributes inherited from parent nodes
        self.children = []
        self.validate()

    def walk(self):
        n_reps = 0
        for repeat in range(self.repeats):
            n_reps += 1
            yield (n_reps, self)
            for child in self.children:
                yield from child.walk()

    @property
    def n_steps(self):
        if type(self) == BaseNode:
            return self.repeats * sum(child.n_steps for child in self.children)
        else:
            return self.repeats

    def __repr__(self):
        return f'<{self.__class__.__name__} "{self.name}" x{self.repeats}>'

    def __str__(self, indent=0):
        """
        Print a tree representation of this node and all children

        :param indent: Recursively increase the indent for child nodes
        :return:
        """
        s = "\t" * indent + self.__repr__()
        for child in self.children:
            s += "\n" + child.__str__(indent=indent + 1)
        return s

    @property
    def attributes(self):
        return sc.mergedicts(self.context, self.instructions)

    def __getitem__(self, item):
        # Directly index the Node to extract attributes without merging the dictionaries every time
        if item in self.instructions:
            return self.instructions[item]
        elif item in self.context:
            return self.context[item]
        else:
            raise KeyError(item)

    def __setitem__(self, key, value):
        self.instructions[key] = value

    def __contains__(self, item):
        return item in self.instructions or item in self.context

    @property
    def repeats(self):
        # Although repeats may be part of the context, we only repeat a node if the instructions requested a repeat
        # i.e., repeats are not inherited
        if isinstance(self.instructions, dict) and "repeats" in self.instructions:
            return self.instructions["repeats"]
        else:
            return 1

    def validate(self):
        """
        Validate/sanitize contents of this node

        If the node isn't valid, an error should be raised
        """
        return

    def apply(self, project: at.Project, parset: at.ParameterSet, savedir, *args, **kwargs) -> ParameterSet:
        """
        Perform the action associated with this node
        """
        return parset


class SectionNode(BaseNode):
    """
    A section node is a special kind of node, that contains other nodes
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.children = self.make_children()
        self.validate()

    def make_children(self):

        children = []

        # Remove any keys in the instructions that correspond to named nodes
        # These should be used as instructions for the child node, rather than
        # forming part of the context that is passed down to all children
        named_nodes = _get_named_nodes()
        step_instructions = {k: v for k, v in self.instructions.items() if isinstance(v, dict) or k in named_nodes}
        for k in step_instructions:
            del self.instructions[k]

        # Create the child nodes
        for name, instructions in step_instructions.items():
            children.append(build(instructions, self.attributes, name))

        return children


class CalibrationNode(BaseNode):

    # Order for list of adjustable parameters and default values
    adj_defaults = {
        "lower_bound": 0.1,
        "upper_bound": 10.0,
        "starting_y_factor": None,
    }

    # Order for list of measurable parameters and default values
    meas_defaults = {
        "weight": 1.0,
        "metric": "fractional",
        "cal_start": -np.inf,
        "cal_end": np.inf,
    }

    @staticmethod
    def parse_list(l, defaults):
        # Routine to parse list of arguments into a dictionary of values
        d = {}
        # convert number strings back to numerical values
        for i, e in enumerate(l.copy()):
            try:
                l[i] = float(e)
            except ValueError:
                pass

        for k, v in zip(list(defaults.keys())[: len(l)], l):
            d[k] = v
        return d

    def validate(self):
        """
        Pre-parse calibration inputs
        """

        def separate_keys(keys_str: str) -> list:
            """
            Separate inputs that kave been defined together as one key in the YAML file but actually represent multiple
            parameters.
            :param str keys_str: Unprocessed input key from the YAML file.
            :return list : A list of strings, each of which represents a single key.
            """
            in_brackets = False
            brackets_str = ""
            nobrackets_str = ""
            separated_keys = []

            for ch in keys_str:
                if ch == "(":
                    in_brackets = True
                    continue
                elif ch == ")":
                    in_brackets = False
                    separated_keys.append(brackets_str)
                    brackets_str = ""
                    continue

                if in_brackets:
                    brackets_str += ch
                else:
                    if ch == ",":
                        if nobrackets_str == " " or nobrackets_str == "":
                            nobrackets_str = ""
                            continue
                        else:
                            separated_keys.append(nobrackets_str)
                            nobrackets_str = ""
                    else:
                        nobrackets_str += ch

            if nobrackets_str != "" and nobrackets_str != " ":
                separated_keys.append(nobrackets_str)

            return [x.strip() for x in separated_keys]

        def process_key(key: str) -> tuple:
            """
            Sanitize the key name, separating the parameter codename from the optional population name/s.
            :par str key: Key representing an adjustable or measurable parameter. It can also contain one or two
            population names (two in the case of a transfer), separated from the parameter name by a comma.
            :returns: A tuple of the parameter codename and population name/s. Population defaults to None
            if not specified.

            EXAMPLES:
                INPUT: 'b_rate'
                OUTPUT: (b_rate, None)

                INPUT: 'b_rate, 0-4'
                OUTPUT: (b_rate, 0-4)

                INPUT: 'aging, 0-4, 5-14'
                OUTPUT: (aging, 0-4, 5-14)
            """
            if "," in key:
                return tuple([x.strip() for x in key.strip("() ").split(",") if x])
            else:
                return (key.strip(), None)

        def process_list(l: list) -> (tuple, list):
            """
            Process list-format inputs and separate them into a key (par_name, pop_name tuple) and a value (list of
            parameter settings).
            @param l: List representing the settings for one parameter.
            @return:    tuple key: Tuple of the parameter codename and the population (default None).
                        list value: List of calibration settings for this parameter.

            EXAMPLES:
                INPUT: [b_rate, 0.1, 10]
                OUTPUT: (b_rate, None) ; [0.1, 10]

                INPUT: [(b_rate, 0-4), 0.1, 10, 1.5]
                OUTPUT: (b_rate, 0-4) ; [0.1, 10, 1.5]
            """
            if len(l) == 1:
                # if the list is already just one string, return that string as key with None pop and vals
                return (l[0].strip("() "), None), None
            elif "(" in str(l):
                # separate out the parenthesis contents as the par/pop/s,
                # then output the key (par, pop tuple) and value

                # process keys
                s = str(l).strip("[] ").replace("'", "")
                s1 = re.findall(r"\(.*?\)", s)
                key = process_key(s1[0].replace("(", "").replace(")", ""))

                # process values/settings
                value = s.replace(s1[0], "").strip(", ").split(",")
                value = [x.strip(", ") for x in value if x]
                return key, value
            else:
                key = process_key(l[0])
                value = l[1:]
                return key, value

        def process_inputs(inputs, defaults: dict) -> dict:
            """
            Process adjustables and measurables, which can be specified as a string, list or nested dict representation.
            * In string representation, only the parameter name is specified, and the default settings are used.
            * In list representation, the input is a list of lists, where the first item in each list is the parameter
            (with optional population) and the remaining items are the supported arguments for the input type, in the
            order defined by the defaults dictionary.
            *In dict representation, the key is the quantity with optional population, and the value can either be
            a list (in the order defined by the dictionary) or a dictionary explicitly naming the inputs.

            This function returns a flat dictionary with {(quantity, pop_name):{argument:value}} e.g., {('b_rate','0-5'):{'lower_bound':0.5}}.
            In the dict representation, the key can be a comma separated list of quantities with optional values e.g.,
            'b_rate 0-5, d_rate'.
            In the list representation, multiple quantities are not supported (as a comma is already used to separate
            the arguments), but multiple lists (one for each quantity) can be provided.
            #TODO could add examples
            """
            out = {}

            if sc.isstring(inputs):
                # Support a comma separated string with "quantity pop" specifications of adjustables and measurables
                # In this case, default values should be used for all other items. Proceed by splitting into a list
                inputs = inputs.split(",")

            if isinstance(inputs, (tuple, list)):
                for l in inputs:
                    l = sc.promotetolist(l)
                    keyspops, v = process_list(l)

                    # process key
                    if len(keyspops) == 2:
                        key, pop_name = keyspops
                    else:
                        assert len(keyspops) == 3, f"Number of populations must be 0, 1 or 2."
                        key = f"{keyspops[0]}_from_{keyspops[1]}"
                        pop_name = keyspops[2]

                    # process value
                    if v is None:
                        value = defaults
                    else:
                        value = self.parse_list(v, defaults)

                    out[key, pop_name] = sc.mergedicts(out.get((key, pop_name), {}), value)

            elif isinstance(inputs, dict):
                for keys, v in inputs.items():

                    separated_keys = separate_keys(keys)
                    for key in separated_keys:
                        # separate par name from pop name
                        keyspops = process_key(key.strip())
                        if len(keyspops) == 2:
                            key, pop_name = keyspops
                        else:
                            assert len(keyspops) == 3, f"Number of populations must be 0, 1 or 2."
                            key = f"{keyspops[0]}_from_{keyspops[1]}"
                            pop_name = keyspops[2]

                        # process values
                        if isinstance(v, (tuple, list)):
                            value = self.parse_list(v, defaults)
                        elif v is None:
                            value = defaults
                        else:
                            value = v.copy()

                        # add keys and values to outputs dict
                        out[key, pop_name] = sc.mergedicts(out.get((key, pop_name), {}), value)
            return out

        self["adjustables"] = process_inputs(self["adjustables"], self.adj_defaults)
        self["measurables"] = process_inputs(self["measurables"], self.meas_defaults)

        def check_optional_number(key, v, defaults):
            if key in v and v[key] is not None:
                if not sc.isnumber(v[key], isnan=False):
                    raise TypeError(f"Adjustable argument {key} needs to be a number or None (defaults to {defaults[key]}). Provided value: {v[key]} ")

        # Validate adjustables
        assert len(self["adjustables"]) > 0, f"Cannot calibrate with no adjustables for calibration section {self.name}"
        for (quantity, pop_name), v in self["adjustables"].items():
            assert "pop_name" not in v, f'Setting the population name through "pop_name: {v["pop_name"]}" is not supported. Instead, the name of the adjustable quantity should include the population name ("{quantity} {v["pop_name"]}")'
            assert isinstance(quantity, str), f"Adjustable codename {quantity} needs to be a string"
            assert pop_name is None or isinstance(pop_name, str), f"Adjustable population {pop_name} needs to be a string or None (defaults to all populations for that parameter)"
            check_optional_number("lower_bound", v, self.adj_defaults)
            check_optional_number("upper_bound", v, self.adj_defaults)
            check_optional_number("starting_y_factor", v, self.adj_defaults)

        # Validate measurables
        assert len(self["measurables"]) > 0, f"Cannot calibrate with no measurables for calibration section {self.name}"
        for (quantity, pop_name), v in self["measurables"].items():
            assert isinstance(quantity, str), f"Measurable codename {quantity} needs to be a string"
            assert pop_name is None or isinstance(pop_name, str), f"Adjustable population {pop_name} needs to be a string or None (defaults to all populations for that parameter)"
            assert "metric" not in v or v["metric"] is None or isinstance(v["metric"], str), f"Measurable metric {v['metric']} needs to be a number or None (defaults to 'fractional')"

            check_optional_number("weight", v, self.meas_defaults)
            check_optional_number("cal_start", v, self.meas_defaults)
            check_optional_number("cal_end", v, self.meas_defaults)

    def apply(self, project: at.Project, parset: at.ParameterSet, n: int, *args, quiet=False, compare_results=False, **kwargs) -> ParameterSet:

        step_name = self.name
        attributes = self.attributes

        at.logger.info(f"Calibrating adjustable(s) {set([adj[0] for adj in attributes['adjustables']])} to match measurable(s) {set([mea[0] for mea in attributes['measurables']])}...")

        # Expand adjustables
        adjustables = {}
        par_names = {x[0] for x in attributes["adjustables"]}.intersection(x.name for x in parset.all_pars())
        pop_names = {x[1] for x in attributes["adjustables"]}.intersection({*parset.pop_names} | {"all", None})

        adj_defaults = {k: self.attributes[k] if k in self.attributes else self.adj_defaults[k] for k in self.adj_defaults}

        for par_name, pop_name in attributes["adjustables"]:

            if par_name not in par_names:
                at.logger.warning(f"Extra YAML adjustable parameter '{par_name}' does not exist in this project's framework/parset and will be ignored")
                continue
            elif pop_name not in pop_names:
                at.logger.warning(f"Extra YAML adjustable population '{pop_name}' does not exist in this project's databook and will be ignored")
                continue

            if pop_name is None:
                pops = parset.pop_names
            else:
                pops = sc.promotetolist(pop_name)

            for pop in pops:
                d = sc.mergedicts(adj_defaults, attributes["adjustables"].get((par_name, None), None), attributes["adjustables"].get((par_name, pop), None))
                adjustables[(par_name, pop)] = (d["lower_bound"], d["upper_bound"], d["starting_y_factor"])
        adjustables = [(*k, *v) for k, v in adjustables.items()]

        # Expand measurables
        measurables = {}
        par_names = {x[0] for x in attributes["measurables"]}.intersection(x.name for x in parset.all_pars())  # TODO: This is probably OK for now but will need to validate that pars have databook entries in the future
        pop_names = {x[1] for x in attributes["measurables"]}.intersection({*parset.pop_names} | {None})

        meas_defaults = {k: self.attributes[k] if k in self.attributes else self.meas_defaults[k] for k in self.meas_defaults}

        for par_name, pop_name in attributes["measurables"]:

            if par_name not in par_names:
                at.logger.warning(f"Extra YAML measurable variable '{par_name}' does not exist in this project's framework and will be ignored")
                continue
            elif pop_name not in pop_names:
                if not (pop_name.lower() == "total" and pop_name.lower() in {x.lower() for x in project.data.tdve[par_name].ts.keys()}):
                    at.logger.warning(f"Extra YAML measurable population '{pop_name}' does not exist in this project's databook and will be ignored")
                    continue

            if pop_name is None:
                pops = parset.pop_names
            else:
                pops = sc.promotetolist(pop_name)

            for pop in pops:
                d = sc.mergedicts(meas_defaults, attributes["measurables"].get((par_name, None), None), attributes["measurables"].get((par_name, pop), None))
                measurables[(par_name, pop)] = (d["weight"], d["metric"], d["cal_start"], d["cal_end"])
        measurables = [(*k, *v) for k, v in measurables.items()]

        # Calibration
        if len(adjustables):
            # note: attributes = instructions + context
            kwargs = sc.mergedicts(self.attributes, kwargs)

            del kwargs["adjustables"]  # supplied via the adjustables variable
            del kwargs["measurables"]  # supplied via the measurables variable

            if "repeats" in kwargs:
                del kwargs["repeats"]

            if quiet:
                with at.Quiet(show_warnings=False):
                    new_cal_parset = at.calibrate(project, parset, adjustables, measurables, **kwargs)
            else:
                new_cal_parset = at.calibrate(project, parset, adjustables, measurables, **kwargs)
        else:
            new_cal_parset = parset

        at.logger.info(f'Completed "{step_name}"...')
        made_changes = False

        for par, pop, *_ in adjustables:

            if pop == "all":
                old = parset.get_par(par).meta_y_factor
                new = new_cal_parset.get_par(par).meta_y_factor
            else:
                old = parset.get_par(par).y_factor[pop]
                new = new_cal_parset.get_par(par).y_factor[pop]

            if new != old:
                at.logger.info(f"...adjusted the y-factor for {par} in {pop} from {old} to {new}")
                made_changes = True
            else:
                at.logger.debug(f"...did NOT adjust the y-factor for {par} in {pop} from {old} to {new}")

        if not made_changes:
            at.logger.info(f"...made no changes!")

        if compare_results:
            base_res = project.run_sim(parset=parset)
            cal_res = project.run_sim(parset=new_cal_parset)
            for par_name in [par_measure[0] for par_measure in measurables]:
                base_rms_error = 0
                cal_rms_error = 0
                for pop in parset.pars[par_name].ts.keys():
                    for time_par_ind, time_value in enumerate(parset.pars[par_name].ts[pop].t):
                        data_time_val = parset.pars[par_name].ts[pop].vals[time_par_ind]
                        base_res_time_ind = list(base_res.get_variable(par_name, pop)[0].t).index(time_value)
                        base_time_val = base_res.get_variable(par_name, pop)[0].vals[base_res_time_ind]
                        cal_res_time_ind = list(cal_res.get_variable(par_name, pop)[0].t).index(time_value)  # probably redundant as they *should* be the same
                        cal_time_val = cal_res.get_variable(par_name, pop)[0].vals[cal_res_time_ind]

                        base_rms_error += (data_time_val - base_time_val) ** 2
                        cal_rms_error += (data_time_val - cal_time_val) ** 2

                        sf = at.get_sigfigs_necessary(base_time_val, cal_time_val)
                        at.logger.info(f"...for parameter {par_name} and population {pop} at time {time_value} the data value was {sc.sigfig(data_time_val, sf)}, the baseline value was {sc.sigfig(base_time_val, sf)}, and the calibrated value was {sc.sigfig(cal_time_val, sf)}.")

                base_rms_error = base_rms_error**0.5
                cal_rms_error = cal_rms_error**0.5
                sf = at.get_sigfigs_necessary(base_rms_error, cal_rms_error)
                at.logger.info(f"...RMS error for parameter {par_name} has changed from baseline {sc.sigfig(base_rms_error, sf)} to calibrated {sc.sigfig(cal_rms_error, sf)}")

        return new_cal_parset


class InitializationNode(BaseNode):
    _name = "set_initialization"

    def __init__(self, instructions, context, name):
        new_instructions = {}

        if isinstance(instructions, dict):
            new_instructions.update(instructions)
        elif type(instructions) is int:
            new_instructions.update({"init_year": instructions})
        elif isinstance(instructions, (tuple, list)):
            sc.promotetolist(instructions)
            new_instructions.update({"init_year": instructions[0]})
            if len(instructions) > 1:
                new_instructions.update({"constant_parset": instructions[1]})

        super().__init__(new_instructions, context, name)

    def validate(self):
        assert "init_year" in self, f"Initialization year must be specified"
        assert sc.isnumber(self["init_year"]), f'Initialization year {self["init_year"]} must be numeric.'
        if "constant_parset" in self:
            assert isinstance(self["constant_parset"], int), f'Constant parset (optional) {self["constant_parset"]} must be numeric (boolean, or int to specify constant parset year).'

    def apply(self, project: at.Project, parset: at.ParameterSet, n: int, *args, **kwargs) -> ParameterSet:
        p2 = sc.dcp(parset)
        if "constant_parset" in self:
            if self["constant_parset"] == False:
                pass
            elif self["constant_parset"] == True:
                p2 = parset.make_constant(year=project.settings.sim_start)
            elif sc.isnumber(self["constant_parset"]):  # constant parset year was provided
                p2 = parset.make_constant(year=self["constant_parset"])
        new_settings = sc.dcp(project.settings)
        new_settings.update_time_vector(end=self["init_year"])
        res = at.run_model(settings=new_settings, framework=project.framework, parset=p2)
        parset.set_initialization(res, self["init_year"])
        return parset


class ClearInitializationNode(BaseNode):
    _name = "clear_initialization"

    def __init__(self, instructions, context, name):
        super().__init__(instructions=None, context=context, name=name)

    def apply(self, project: at.Project, parset: at.ParameterSet, n: int, *args, **kwargs) -> ParameterSet:
        parset.initialization = None
        return parset


class SaveCalibrationNode(BaseNode):
    """
    Block in YAML file with "save calibration: <file name>"
    """

    _name = "save_calibration"

    def __init__(self, instructions, context, name):
        if not isinstance(instructions, dict):
            instructions = {"fname": instructions}
        super().__init__(instructions, context, name)

    def validate(self):
        assert self["fname"] is not None, 'A "save calibration" node must have a file name explicitly specified'

    def apply(self, project: at.Project, parset: at.ParameterSet, savedir=None, *args, **kwargs) -> ParameterSet:
        parset.save_calibration(savedir / self["fname"])
        return parset



# Test that parameter derivatives work
# Indirectly test that frameworks with no compartments work

import numpy as np
import atomica as at
import os

testdir = at.parent_dir()  # Must be relative to current file to work with tox


def test_table_parsing():
    F = at.ProjectFramework(testdir / "test_table_parsing.xlsx")


def test_table_parsing2():
    F = at.ProjectFramework(testdir / "test_table_parsing2.xlsx")

    assert len(F.sheets["extra 0"]) == 2
    assert len(F.sheets["extra 0"][0]) == 2
    assert len(F.sheets["extra 0"][1]) == 2

    assert len(F.sheets["extra 1"]) == 1
    assert F.sheets["extra 1"][0].empty
    assert F.sheets["extra 1"][0].columns.tolist() == ["heading 1", "heading 2"]

    assert len(F.sheets["extra 2"]) == 3
    assert len(F.sheets["extra 2"][0]) == 0
    assert len(F.sheets["extra 2"][1]) == 1
    assert len(F.sheets["extra 2"][2]) == 0

    assert len(F.sheets["extra 3"]) == 1
    assert len(F.sheets["extra 3"][0]) == 1

    assert len(F.sheets["extra 4"]) == 2
    assert len(F.sheets["extra 4"][0]) == 0
    assert len(F.sheets["extra 4"][1]) == 1


if __name__ == "__main__":
    # test_table_parsing()
    test_table_parsing2()



# Test that adding an extra population in the data for plotting is OK

import atomica as at


def test_extra_data_pop():
    testdir = at.parent_dir()

    F_path = testdir / "framework_sir_dynamic.xlsx"
    D_path = testdir / "databook_sir_dynamic_extra.xlsx"

    P = at.Project(framework=F_path, databook=D_path, do_run=True)

    d = at.PlotData(P.results[0], "sus")
    at.plot_series(d, data=P.data)

    d.series[0].data_pop = "national"
    at.plot_series(d, data=P.data)


if __name__ == "__main__":
    test_extra_data_pop()



"""
Implements automatic calibration

This module defines the :func:`calibrate` function, which is the entry-point for
automatic calibration

"""

import numpy as np
import sciris as sc
from .model import BadInitialization
from .system import logger
from .parameters import ParameterSet
import logging
import atomica

__all__ = ["calibrate"]

# TODO: Determine whether this is necessary.
calibration_settings = dict()
calibration_settings["tolerance"] = 1e-6


def _update_parset(parset, y_factors, pars_to_adjust) -> None:
    """

    :param parset: a ParameterSet object
    :param y_factors: Array with as many elements as pars_to_adjust
    :param pars_to_adjust: Array of tuples (par_name,pop_name,...) with special value pop='all' supported to set meta factor
                           Must have as many elements as y_factors. pop=None is not allowed - it must be converted
                           to a full list of pops previously (in perform_autofit)
    :return: None (parset is modified in-place)
    """

    for i, (par_name, pop_name, *_) in enumerate(pars_to_adjust):
        par = parset.get_par(par_name)
        if pop_name.lower() == "all":
            par.meta_y_factor = y_factors[i]
        else:
            par.y_factor[pop_name] = y_factors[i]


def _calculate_objective(y_factors, pars_to_adjust, output_quantities, parset, project, *args, **kwargs) -> float:
    """
    Run the model for a given set of y-factors and return the objective/goodness-of-fit

    Additional extra arguments will be ignored but will not raise an error.

    :param y_factors: array of y-factors to apply to specified output_quantities
    :param pars_to_adjust: list of tuples (par_name,pop_name,...) recognized by parset.update()
    :param output_quantities: a tuple containing (pop, var, weight, metric, start_year, end_year) - start year and end year are inclusive
    :param parset:
    :param project:
    :return: The value of the objective function defined by the output_quantities
    """

    _update_parset(parset, y_factors, pars_to_adjust)

    try:
        result = project.run_sim(parset=parset, store_results=False)
    except BadInitialization:  # If the proposed parameters lead to invalid initial compartment sizes
        return np.inf

    objective = 0.0

    for var_label, pop_name, weight, metric, start_year, end_year in output_quantities:

        target = project.data.get_ts(var_label, pop_name)  # This is the TimeSeries with the data for the requested quantity
        if target is None:
            continue
        if not target.has_time_data:  # Only use this output quantity if the user entered time-specific data
            continue

        if pop_name.lower() == "total":
            var = atomica.PlotData(result, outputs=var_label, pops="total", project=project)
        else:
            var = result.model.get_pop(pop_name).get_variable(var_label)

        data_t, data_v = target.get_arrays()
        inds = (data_t >= start_year) & (data_t <= end_year)
        if np.count_nonzero(inds) == 0:
            # If no time points remain after filtering down to the time points the user requested
            logger.info(f"No data points remaining after filtering down to requested time period. Skipping...")
            continue
        data_t = data_t[inds]
        data_v = data_v[inds]

        # Interpolate the model outputs onto the data times
        # If there is data outside the range when the model was simulated, don't
        # extrapolate the model outputs
        y = data_v
        if pop_name.lower() == "total":
            y2 = np.interp(data_t, var.series[0].tvec, var.series[0].vals, left=np.nan, right=np.nan)
        else:
            y2 = np.interp(data_t, var[0].t, var[0].vals, left=np.nan, right=np.nan)

        idx = ~np.isnan(y) & ~np.isnan(y2)
        objective += weight * sum(_calculate_fitscore(y[idx], y2[idx], metric))

    return objective


def _get_fitscore_func(metric):
    availfns = globals().copy()
    availfns.update(locals())
    try:
        return availfns.get("_calc_%s" % metric)
    except Exception:
        raise NotImplementedError("No method associated with _calc_%s (calibration.py)" % metric)


def _calculate_fitscore(y_obs, y_fit, metric="meansquare"):
    return _get_fitscore_func(metric)(y_obs, y_fit)


def _calc_meansquare(y_obs, y_fit):
    """
    Calcs the RMS error.

    Note: could also use implementation from sklearn in future ...
    """
    return np.sqrt(((y_fit - y_obs) ** 2).mean())


def _calc_fractional(y_obs, y_fit):
    return np.abs((y_fit - y_obs) / np.clip(y_obs, 1, None))  # Use clipping to handle cases where data has value 0


def _calc_wape(y_obs, y_fit):
    """
    Calculates the weighted absolute percentage error
    """
    return abs(y_fit - y_obs) / (y_obs.mean() + calibration_settings["tolerance"])


def calibrate(project, parset: ParameterSet, pars_to_adjust, output_quantities, max_time=None, method="asd", time_period=(-np.inf, np.inf), **kwargs) -> ParameterSet:
    """
    Run automated calibration

    :param project: A project instance to provide data and sim settings
    :param parset: A :class:`ParameterSet` instance to calibrate
    :param pars_to_adjust: list of tuples, `(par_name, pop_name, lower_bound, upper_bound, initial_value)`
                           the pop name can be None, which will be expanded to all populations
                           relevant to the parameter independently, or 'all' which will instead operate
                           on the meta y factor. To calibrate a transfer, the parameter name should be set to
                           ``'<tranfer_code_name>_from_<from_pop>'`` and then the destination population can be specified
                           as the ``pop_name``. For example, to automatically calibrate an aging transfer 'age' from 5-14
                           to 15-64, the tuple would contain ``pars_to_adjust=[('age_from_5-14','15-64',...)]``
    :param output_quantities: list of tuples, (var_label,pop_name,weight,metric), for use in the objective
                              function. pop_name=None will expand to all pops. pop_name='all' is not supported. The
                              output can optionally contain `(var_label, pop_name, weight, metric, start_year, end_year)`
                              to select a subset of the data for evaluating the objective. The start year and end year
                              specified here will take precedence over the time_period argument. In some cases, it may be desirable
                              to fit to an aggregated total value across populations. In that case, the databook should have
                              an extra row in the TDVE table for a population called "Total". The measurable
                              can then be given pop_name="Total" which will cause the Atomica model outputs to be aggregated over
                              all populations, and the aggregate value compared to the "Total" data. The aggregation methods will
                              be automatically selected depending on units of the quantity (sum for "number" units, average for others).
    :param max_time: If using ASD, the maximum run time
    :param time_period: Tuple of start and end years to use for the objective function. Applies to all outputs unless
                        the output has an explicitly specified start and end year
    :param method: 'asd' or 'pso'. If using 'pso' all upper and lower limits must be finite
    :param kwargs: Dictionary of additional arguments to be passed to the optimization function, e.g. stepsize or pinitial
    :return: A calibrated :class:`ParameterSet`

    """

    # Expand out pop=None in pars_to_adjust
    p2 = []
    for par_tuple in pars_to_adjust:
        if len(par_tuple) == 5:
            initial_value = par_tuple[4]
        else:
            initial_value = None

        if par_tuple[1] is None:  # If the pop name is None
            par = parset.get_par(par_tuple[0])
            for pop_name in par.pops:
                p2.append((par_tuple[0], pop_name, par_tuple[2], par_tuple[3], initial_value))
        else:
            p2.append((par_tuple[0], par_tuple[1], par_tuple[2], par_tuple[3], initial_value))

    pars_to_adjust = p2

    # Expand out pop=None in output_quantities
    outputs = []
    for output_tuple in output_quantities:
        var_label = output_tuple[0]
        pop_name = output_tuple[1]
        weight = output_tuple[2]
        metric = output_tuple[3]
        if len(output_tuple) == 6:
            start_year = output_tuple[4]
            end_year = output_tuple[5]
        else:
            start_year = time_period[0]
            end_year = time_period[1]

        if pop_name is None:  # If the pop name is None
            for pop in project.data.pops.keys():
                outputs.append((var_label, pop, weight, metric, start_year, end_year))
        else:
            outputs.append((var_label, pop_name, weight, metric, start_year, end_year))

    output_quantities = outputs

    x0 = []
    xmin = []
    xmax = []
    filtered_pars_to_adjust = []
    parset = parset.copy()

    for i, x in enumerate(pars_to_adjust):
        par_name, pop_name, scale_min, scale_max, initial_value = x

        par = parset.get_par(par_name)

        # if initial_value has not been explicitly set in the tuple: use y_factor in parset
        if initial_value is None:
            if pop_name == "all":
                initial_value = par.meta_y_factor
            else:
                initial_value = par.y_factor[pop_name]
            # if this value is outside the min and max bounds, make it equal to min or max (whichever is closest)
            # if min == max, this should make the initial value equal to min and max
            initial_value = np.clip(initial_value, scale_min, scale_max)
        else:
            assert (initial_value >= scale_min) and (initial_value <= scale_max), "Initial value is not consistent with the lower/upper bounds"

        # update y_factors in parset
        if pop_name == "all":
            par.meta_y_factor = initial_value
        else:
            par.y_factor[pop_name] = initial_value

        if scale_min != scale_max:
            # Only include the y-factor in the adjustments made in the optimization function if a range
            # of y-factor values is permitted
            filtered_pars_to_adjust.append(x)
            x0.append(initial_value)
            xmin.append(scale_min)
            xmax.append(scale_max)

    args = {
        "project": project,
        "parset": parset,
        "pars_to_adjust": filtered_pars_to_adjust,
        "output_quantities": output_quantities,
    }

    original_sim_end = project.settings.sim_end
    project.settings.sim_end = min(project.data.tvec[-1], original_sim_end)

    if len(filtered_pars_to_adjust) > 0:
        try:
            if method == "asd":
                optim_args = {
                    "stepsize": 0.1,
                    "maxiters": 2000,
                    "sinc": 1.5,
                    "sdec": 2.0,
                    "reltol": 1e-3,
                    "abstol": 1e-6,
                    "xmin": xmin,
                    "xmax": xmax,
                    "maxtime": 60 if max_time is None else max_time,
                    "minval": 0,
                }
                optim_args.update(kwargs)

                if "verbose" not in optim_args:
                    log_level = logger.getEffectiveLevel()
                    if log_level < logging.WARNING:
                        optim_args["verbose"] = 2
                    else:
                        optim_args["verbose"] = 0

                opt_result = sc.asd(_calculate_objective, x0, args, **optim_args)
                x1 = opt_result["x"]
            elif method == "pso":
                import pyswarm

                optim_args = {"maxiter": 3, "lb": xmin, "ub": xmax, "minstep": 1e-3, "debug": True}
                if np.any(~np.isfinite(xmin)) or np.any(~np.isfinite(xmax)):
                    errormsg = "PSO optimization requires finite upper and lower bounds to specify the search domain (i.e. every parameter being adjusted needs to have finite bounds)"
                    raise Exception(errormsg)

                x1, _ = pyswarm.pso(_calculate_objective, kwargs=args, **optim_args)
            else:
                raise Exception("Unrecognized method")
        except Exception as e:
            raise e
        finally:
            project.settings.sim_end = original_sim_end  # Restore the simulation end year

        _update_parset(args["parset"], x1, args["pars_to_adjust"])
    else:
        logger.info("No parameters to adjust provided to the optimisation function. Skipping optimisation...")

    # Log out the commands required for equivalent manual calibration if desired
    for i, x in enumerate(pars_to_adjust):
        par_name = x[0]
        pop_name = x[1]

        if par_name in parset.pars:
            par = args["parset"].pars[par_name]

            if pop_name.lower() == "all":
                logger.debug("parset.get_par('{}').meta_y_factor={:.2f}".format(par_name, par.meta_y_factor))
            else:
                logger.debug("parset.get_par('{}').y_factor['{}']={:.2f}".format(par_name, pop_name, par.y_factor[pop_name]))

        else:
            tokens = par_name.split("_from_")
            par = args["parset"].transfers[tokens[0]][tokens[1]]
            logger.debug("parset.transfers['{}']['{}'].y_factor['{}']={:.2f}".format(tokens[0], tokens[1], pop_name, par.y_factor[pop_name]))

    args["parset"].name = "calibrated_" + args["parset"].name

    return args["parset"]



"""
Implementation of Databook functionality

This module defines the :class:`ProjectData` class, which serves as a
Python-based representation of the Databook, as well as providing methods for
reading Databooks into `ProjectData` instances, and saving `ProjectData` back
to Excel files.

"""

from .utils import TimeSeries
import sciris as sc
from xlsxwriter.utility import xl_rowcol_to_cell as xlrc
from .excel import cell_get_string, standard_formats, read_tables, TimeDependentValuesEntry, TimeDependentConnections, apply_widths, update_widths, validate_category
import xlsxwriter as xw
import io
import numpy as np
from .system import NotFoundError
from . import logger
from .system import FrameworkSettings as FS
from collections import defaultdict
import pandas as pd
import itertools
from .version import version, gitinfo

__all__ = ["InvalidDatabook", "ProjectData"]


class InvalidDatabook(Exception):
    pass


class ProjectData(sc.prettyobj):
    """
    Store project data: class-equivalent of Databooks

    This class is used to load and work with data that is entered in databooks. It
    provides the interface for

    - Loading data
    - Modifying data (values, adding/removing populations etc.
    - Saving modified data
    - Writing new databooks

    To instantiate, the ``ProjectData`` constructor is normally not used. Instead, use
    the static methods

    - ``ProjectData.new()`` to create a new instance/databook given a :class:`ProjectFramework`
    - ``ProjectData.from_spreadsheet()`` to load a databook

    """

    def __init__(self, framework):
        self.pops = sc.odict()  #: This is an odict mapping code_name:{'label':full_name, 'type':pop_type}
        self.transfers = list()  #: This stores a list of :class:`TimeDependentConnections` instances for transfers
        self.interpops = list()  #: This stores a list of :class:`TimeDependentConnections` instances for interactions
        self.tvec = None  #: This is the data's tvec used when instantiating new tables. Not _guaranteed_ to be the same for every TDVE/TDC table
        self.tdve = sc.odict()  #: This is an odict storing :class:`TimeDependentValuesEntry` instances keyed by the code name of the TDVE
        self.tdve_pages = sc.odict()  #: This is an odict mapping worksheet name to an (ordered) list of TDVE code names appearing on that sheet

        # Internal storage used with methods while writing
        self._pop_types = list(framework.pop_types.keys())  #: Store set of valid population types from framework
        self._formats = None  #: Temporary storage for the Excel formatting while writing a databook
        self._book = None  #: Temporary storage for the workbook while writing a databook
        self._references = None  #: Temporary storage for cell references while writing a databook

        self.version = version  #: Current Atomica version
        self.gitinfo = sc.dcp(gitinfo)  #: Atomica Git version information, if being run in a Git repository

    def __setstate__(self, d):
        from .migration import migrate

        self.__dict__ = d
        projectdata = migrate(self)
        self.__dict__ = projectdata.__dict__

    def tables(self):
        """
        Return iterator over all TDVE and TDC tables

        :return: An iterator
        """
        for table in itertools.chain(self.tdve.values(), self.transfers, self.interpops):
            yield table

    @property
    def start_year(self) -> float:
        """
        Return the start year from the databook

        The ProjectData start year is defined as the earliest time point in
        any of the TDVE/TDC tables (noting that it it is possible for the TDVE tables to
        have different time values). This quantity should be used when changing the simulation
        start year, if using all of the data in the databook is desired.

        :return: The earliest year in the databook

        """

        start_year = np.inf
        for table in self.tables():
            if len(table.tvec) and np.amin(table.tvec) < start_year:
                start_year = np.amin(table.tvec)
        return start_year

    @property
    def end_year(self) -> float:
        """
        Return the start year from the databook

        The ProjectData end year is defined as the latest time point in
        any of the TDVE/TDC tables (noting that it it is possible for the TDVE tables to
        have different time values). This quantity should be used when changing the simulation
        end year, if using all of the data in the databook is desired.

        :return: The latest year in the databook

        """

        end_year = -np.inf
        for table in self.tables():
            if len(table.tvec) and np.amax(table.tvec) > end_year:
                end_year = np.amax(table.tvec)
        return end_year

    def change_tvec(self, tvec: np.array) -> None:
        """
        Change the databook years

        This function can be used to change the time vector in all of the TDVE/TDC tables.
        There are two ways to change the time arrays:

        - Setting ``ProjectData.tvec`` directly will only affect newly added tables, and will keep existing tables
          as they are
        - Calling ``ProjectData.change_tvec()`` will modify all existing tables

        Note that the TDVE/TDC tables store time/value pairs sparsely within their :class:`TimeSeries` objects.
        Therefore, changing the time array won't modify any of the data - it will only have an effect the next time
        a databook is written (so typically this method would be called as part of preparing a modified databook).

        :param tvec: A float, list, or array containing time values (in years) for the databook

        """

        self.tvec = sc.promotetoarray(tvec).copy()
        for table in self.tables():
            table.tvec = tvec

    def get_ts(self, name: str, key=None):
        """
        Extract a TimeSeries from a TDVE table or TDC table

        :param name: The code name for the container storing the :class:`TimeSeries`
                    - The code name of a transfer, interaction, or compartment/characteristic/parameter
                    - The name of a transfer parameter instantiated in model.build e.g. 'age_0-4_to_5-14'.
                    this is mainly useful when retrieving data for plotting, where variables are organized according
                    to names like 'age_0-4_to_5-14'
        :param key: Specify the identifier for the :class:`TimeSeries`
                        - If `name` is a comp/charac/par, then key should be a pop name
                        - If `name` is a transfer or interaction, then key should be a tuple (from_pop,to_pop)
                        - If `name` is the name of a model transfer parameter, then `key` should be left as `None`
        :return: A :class:`TimeSeries`, or ``None`` if there were no matches

        Regarding the specification of the key - the same transfer could be specified as

        - ``name='age', key=('0-4','5-14')``
        - ``name='age_0-4_to_5-14', key=None``

        where the former is typically used when working with data and calibrations, and the latter is used in :class:`Model` and
        is therefore encountered on the :class:`Result` and plotting side.

        If retrieving values for a comp/charac/par and the databook contains an entry for 'all' rather
        than specific populations, then the 'all' time series will be returned regardless of the key
        """

        # Exit immediately if the name is not specified
        if not name:
            return None

        # First, check if it's the name of a TDVE
        if name in self.tdve:
            if key in self.tdve[name].ts:
                return self.tdve[name].ts[key]
            elif "all" in self.tdve[name].ts:
                return self.tdve[name].ts["all"]
            elif "All" in self.tdve[name].ts:
                return self.tdve[name].ts["All"]

        # If the key is specified, then the name corresponds to the code name only, and we can just directly
        # use the name and the key. However, if the key is none, then the user has passed in the name as
        # "<code_name>_<from_pop>_to_<to_pop>" e.g., "age_0-4_to_5-14". We need to split this up then into
        # the transfer/interaction name, and the key. Note that the code name may contain underscores
        for tdc in self.transfers + self.interpops:
            if name.startswith(tdc.code_name):
                key = key or tuple(name[len(tdc.code_name) + 1 :].split("_to_"))
                if key in tdc.ts:
                    return tdc.ts[key]
                break  # If we matched the name of the TDC but the requestion populations are wrong, no need to check other TDCs

        return None

    def get_tdve_page(self, code_name) -> str:
        """
        Given a code name for a TDVE quantity, find which page it is on

        :param code_name: The code name for a TDVE quantity
        :return: The sheet that it appears on

        """

        for sheet, content in self.tdve_pages.items():
            if code_name in content:
                return sheet
        else:
            raise NotFoundError('The quantity "%s" does not appear on any TDVE sheets' % (code_name))

    @staticmethod
    def new(framework, tvec, pops, transfers):
        """
        Make a new databook/``ProjectData`` instance

        This method should be used (instead of the standard constructor) to produce a new
        class instance (e.g. if creating a new databook).

        :param framework: A :class:`ProjectFramework` instance
        :param tvec: A scalar, list, or array of times (typically would be generated with ``numpy.arange()``)
        :param pops: A number of populations, or a ``dict`` with either ``{name:label}`` or ``{name:{label:label,type:type}}``. Type defaults
                     to the first population type in the framework
        :param transfers: A number of transfers, or a ``dict`` with either ``{name:label}`` or ``{name:{label:label,type:type}}``.
                     The type defaults to the first population type in the framework. Transfers can only take place between populations of the
                     same type.
        :return: A new :class:`ProjectData` instance

        """

        new_pops = sc.odict()
        default_pop_type = list(framework.pop_types.keys())[0]

        if sc.isnumber(pops):
            for i in range(0, pops):
                new_pops["pop_%d" % (i)] = {"label": "Population %d" % (i), "type": default_pop_type}
        else:
            for code_name, spec in pops.items():
                if sc.isstring(spec):
                    new_pops[code_name] = {"label": spec, "type": default_pop_type}
                else:
                    new_pops[code_name] = spec

        if not new_pops:
            raise InvalidDatabook("A new databook must have at least 1 population")

        new_transfers = sc.odict()
        if sc.isnumber(transfers):
            for i in range(0, transfers):
                new_transfers["transfer_%d" % (i)] = {"label": "Transfer %d" % (i), "type": default_pop_type}
        else:
            for code_name, spec in transfers.items():
                if sc.isstring(spec):
                    new_transfers[code_name] = {"label": spec, "type": default_pop_type}
                else:
                    new_transfers[code_name] = spec

        # Make all of the empty TDVE objects - need to store them by page, and the page information is in the Framework
        data = ProjectData(framework=framework)
        data.tvec = sc.promotetoarray(tvec)
        pages = defaultdict(list)  # This will store {sheet_name:(code_name,databook_order)} which will then get sorted further

        for obj_type, df in zip(["comps", "characs", "pars"], [framework.comps, framework.characs, framework.pars]):
            for _, spec in df.iterrows():
                databook_page = spec.get("databook page")
                if not pd.isna(databook_page):
                    pop_type = spec.get("population type")
                    databook_order = spec.get("databook order")
                    full_name = spec["display name"]
                    default_all = spec["databook default all"] == "y"
                    allowed_units = [framework.get_databook_units(full_name)]

                    if pd.isna(databook_order):
                        order = np.inf
                    else:
                        order = databook_order
                    pages[databook_page].append((spec.name, order))
                    data.tdve[spec.name] = TimeDependentValuesEntry(full_name, data.tvec, allowed_units=allowed_units, comment=spec["guidance"], pop_type=pop_type, default_all=default_all)
                    data.tdve[spec.name].write_units = True
                    data.tdve[spec.name].write_uncertainty = True
                    if obj_type == "pars":
                        data.tdve[spec.name].write_assumption = True
                        if spec["timed"] == "y":
                            data.tdve[spec.name].tvec = []  # If parameter is timed, don't show any years
                            data.tdve[spec.name].write_uncertainty = False  # Don't show uncertainty for timed parameters. In theory users could manually add the column and sample over it, but because the duration is rounded to the timestep, it's likely to have confusing stepped effects
                    data.tdve[spec.name].pop_type = pop_type

                    if default_all:
                        # add_pop normally adds TDVE rows, but it won't operate on any TDVEs that default to 'All' so we need to add the 'All' rows here
                        data.tdve[spec.name].ts["All"] = TimeSeries(units=allowed_units[0])

        # Now convert pages to full names and sort them into the correct order
        for _, spec in framework.sheets["databook pages"][0].iterrows():

            if spec["datasheet code name"] in pages:
                pages[spec["datasheet code name"]].sort(key=lambda x: x[1])
                data.tdve_pages[spec["datasheet title"]] = [x[0] for x in pages[spec["datasheet code name"]]]
            else:
                data.tdve_pages[spec["datasheet title"]] = list()

        # Now, proceed to add pops, transfers, and interactions
        for code_name, spec in new_pops.items():
            data.add_pop(code_name, spec["label"], pop_type=spec["type"])

        for code_name, spec in new_transfers.items():
            data.add_transfer(code_name, spec["label"], pop_type=spec["type"])

        for _, spec in framework.interactions.iterrows():
            interpop = data.add_interaction(spec.name, spec["display name"], from_pop_type=spec["from population type"], to_pop_type=spec["to population type"])
            if "default value" in spec and np.isfinite(spec["default value"]):
                for from_pop in interpop.from_pops:
                    for to_pop in interpop.to_pops:
                        ts = TimeSeries(units=interpop.allowed_units[0])
                        ts.insert(None, spec["default value"])
                        interpop.ts[(from_pop, to_pop)] = ts
                        interpop.ts_attributes["Provenance"][(from_pop, to_pop)] = spec["provenance"] if "provenance" in spec else FS.DEFAULT_PROVENANCE

        # Finally, insert parameter and characteristic default values
        for df in [framework.comps, framework.characs, framework.pars]:
            for _, spec in df.iterrows():
                # In order to write a default value
                # - The default value should be present and not None
                # - The quantity should appear in the databook
                if "default value" in spec and np.isfinite(spec["default value"]) and not pd.isna(spec["databook page"]):
                    tdve = data.tdve[spec.name]
                    for key, ts in tdve.ts.items():
                        ts.insert(None, spec["default value"])
                        tdve.ts_attributes["Provenance"][key] = spec["provenance"] if "provenance" in spec else FS.DEFAULT_PROVENANCE

        return data

    @staticmethod
    def from_spreadsheet(spreadsheet, framework):
        """
        Construct ProjectData from spreadsheet

        The framework is needed because the databook does not read in or otherwise store
            - The valid units for quantities
            - Which population type is associated with TDVE tables

        :param spreadsheet: The name of a spreadsheet, or a `sc.Spreadsheet`
        :param framework: A :class:`ProjectFramework` instance
        :return: A new :class:`ProjectData` instance

        """

        # Basically the strategy is going to be
        # 1. Read in all of the stuff - pops, transfers, interpops can be directly added to Data
        # 2. Read in all the other TDVE content, and then store it in the data specs according to the variable type defined in the Framework
        # e.g. the fact that 'Alive' is a Characteristic is stored in the Framework and Data but not in the Databook. So for example, we read in
        # a TDVE table called 'Alive', but it needs to be stored in data.specs['charac']['ch_alive'] and the 'charac' and 'ch_alive' are only available in the Framework

        import openpyxl

        self = ProjectData(framework=framework)

        if not isinstance(spreadsheet, sc.Spreadsheet):
            spreadsheet = sc.Spreadsheet(spreadsheet)

        workbook = openpyxl.load_workbook(spreadsheet.tofile(), read_only=True, data_only=True)  # Load in read-only mode for performance, since we don't parse comments etc.
        validate_category(workbook, "atomica:databook")

        # These sheets are optional - if none of these are provided in the databook
        # then they will remain empty
        self.transfers = list()
        self.interpops = list()

        for sheet in workbook.worksheets:

            if sheet.title.startswith("#ignore"):
                continue

            if sheet.title == "Population Definitions":
                try:
                    self._read_pops(sheet)
                except Exception as e:
                    message = 'An error was detected on the "Population Definitions" sheet'
                    raise InvalidDatabook("%s -> %s" % (message, e)) from e
            elif sheet.title == "Transfers":
                try:
                    self._read_transfers(sheet)
                except Exception as e:
                    message = 'An error was detected on the "Transfers" sheet'
                    raise InvalidDatabook("%s -> %s" % (message, e)) from e
            elif sheet.title == "Interactions":
                try:
                    self._read_interpops(sheet)
                except Exception as e:
                    message = 'An error was detected on the "Interactions" sheet'
                    raise InvalidDatabook("%s -> %s" % (message, e)) from e
            elif sheet.title == "Metadata":
                continue
            else:
                self.tdve_pages[sheet.title] = []
                tables, start_rows = read_tables(sheet)
                for table, start_row in zip(tables, start_rows):

                    try:
                        tdve = TimeDependentValuesEntry.from_rows(table)
                    except Exception as e:
                        message = 'Error on sheet "%s" while trying to read a TDVE table starting on row %d' % (sheet.title, start_row)
                        raise InvalidDatabook("%s -> %s" % (message, e)) from e

                    # If the TDVE is not in the Framework, that's a critical stop error, because the framework needs to at least declare
                    # what kind of variable this is - otherwise, we don't know the allowed units and cannot write the databook back properly
                    try:
                        spec, item_type = framework.get_variable(tdve.name)
                    except NotFoundError:
                        message = 'Error on sheet "%s" while reading TDVE table "%s" (row %d). The variable was not found in the Framework' % (sheet.title, tdve.name, start_row)
                        raise InvalidDatabook(message)

                    code_name = spec.name
                    tdve.allowed_units = [framework.get_databook_units(code_name)]
                    tdve.pop_type = spec["population type"]

                    # Migrate the units (20181114)
                    # All TimeSeries instances in databook TDVE tables should have the same units as the allowed units
                    # However, if the user entered something that is wrong, we need to keep it and alert them during validation
                    # Therefore, we can migrate as long as the _old_ units made sense
                    for ts in tdve.ts.values():
                        if ts.units != tdve.allowed_units[0]:
                            if not ts.units or ts.units.strip().lower() == tdve.allowed_units[0].strip().split()[0].strip().lower():
                                ts.units = tdve.allowed_units[0]

                    if not spec["databook page"]:
                        # Note that if the parameter doesn't have a databook page and the framework is valid, then the parameter must have a function. Therefore,
                        # if data is also read in, it will not change the simulation outputs and would only be used for calibration/validation
                        logger.warning('A TDVE table for "%s" (%s) was read in and data will be available for calibration, but the Framework did not mark this quantity as appearing in the databook', tdve.name, code_name)
                    tdve.comment = spec["guidance"]

                    if code_name in self.tdve:
                        raise InvalidDatabook('A TDVE table for "%s" (%s) appears more than once in the databook. The first table was on sheet "%s" and the first duplicate table is on sheet "%s" starting on row %d' % (tdve.name, code_name, [k for k, v in self.tdve_pages.items() if code_name in v][0], sheet.title, start_row))

                    self.tdve[code_name] = tdve
                    # Store the TDVE on the page it was actually on, rather than the one in the framework. Then, if users move anything around, the change will persist
                    self.tdve_pages[sheet.title].append(code_name)

        # Check that transfer and interaction names are unique
        _interactions = {}
        for interaction in self.transfers + self.interpops:
            if interaction.code_name not in _interactions:
                _interactions[interaction.code_name] = interaction
            else:
                raise InvalidDatabook(f'Duplicate name - {interaction.type.title()} "{interaction.code_name}" has the same name as an existing {_interactions[interaction.code_name].type}"')

        tvals = set()
        for table in self.tables():
            tvals.update(table.tvec)
        self.tvec = np.array(sorted(tvals))

        return self

    def validate(self, framework) -> bool:
        """
        Check if the ProjectData instance can be used to run simulations

        A databook can be 'valid' in two senses

        - The Excel file adheres to the correct syntax and it can be parsed into a ProjectData object
        - The resulting ProjectData object contains sufficient information to run a simulation

        Sometimes it is desirable for ProjectData to be valid in one sense rather than the other. For example,
        in order to run a simulation, the ProjectData needs to contain at least one value for every TDVE table.
        However, the TDVE table does _not_ need to contain values if all we want to do is add another key pop
        Thus, the first stage of validation is the ProjectData constructor - if that runs, then users can
        access methods like 'add_pop','remove_transfer' etc.

        On the other hand, to actually run a simulation, the _contents_ of the databook need to satisfy various conditions
        These tests are implemented here. The typical workflow would be that ProjectData.validate() should be used
        if a simulation is going to be run. In the first instance, this can be done in `Project.load_databook` but
        the FE might want to perform this check at a different point if the databook manipulation methods e.g.
        `add_pop` are going to be exposed in the interface

        This function throws an informative error if there are any problems identified or otherwise returns True

        :param framework: A :class:`ProjectFramework` instance to validate the data against
        :return: True if ProjectData is valid. An error will be raised otherwise

        """

        # Make sure that all of the quantities the Framework says we should read in have been read in, and that
        # those quantities all have some data values associated with them
        for pop in self.pops.values():
            if pd.isna(pop["type"]):
                pop["type"] = self._pop_types[0]
            assert pop["type"] in self._pop_types, 'Error in population "%s": population type "%s" not found in framework. If the framework defines a non-default population type, then it must be explicitly specified in databooks and program books.' % (pop["label"], pop["type"])

        for obj_type, df in zip(["comps", "characs", "pars"], [framework.comps, framework.characs, framework.pars]):
            for spec_name, spec in zip(df.index, df.to_dict(orient="records")):
                if spec_name in self.pops:
                    raise InvalidDatabook('Code name "%s" has been used for both a population and a framework quantity - population names must be unique' % (spec_name))

                if not pd.isna(spec["databook page"]):
                    if spec_name not in self.tdve:
                        if not np.isfinite(spec["default value"]):
                            raise InvalidDatabook('The databook did not contain a required TDVE table named "%s" (code name "%s")' % (spec["display name"], spec_name))
                        else:
                            logger.warning('TDVE table "%s" (code name "%s") is missing from the databook. Using default values from the framework' % (spec["display name"], spec_name))
                            units = framework.get_databook_units(spec_name)
                            self.tdve[spec_name] = TimeDependentValuesEntry(spec["display name"], self.tvec.copy(), allowed_units=[units], comment=spec["guidance"], pop_type=spec["population type"])
                            for pop in self.pops.keys():
                                self.tdve[spec_name].ts[pop] = TimeSeries(assumption=spec["default value"], units=units)
                            tdve_page = framework.sheets["databook pages"][0][framework.sheets["databook pages"][0]["datasheet code name"] == spec["databook page"]]["datasheet title"].values[0]
                            for existing in self.tdve_pages.keys():
                                if existing.lower() == tdve_page.lower():
                                    self.tdve_pages[existing].append(spec_name)
                                    break
                            else:
                                self.tdve_pages[tdve_page] = [spec_name]
                    else:
                        framework_units = framework.get_databook_units(spec_name)  # Get the expected databook units
                        tdve = self.tdve[spec_name]
                        tdve_sheet = self.get_tdve_page(spec_name)
                        location = 'Error in TDVE table "%s" on sheet "%s"' % (tdve.name, tdve_sheet)
                        assert tdve.pop_type in self._pop_types, '%s. Population type "%s" did not match any in the framework' % (location, tdve.pop_type)
                        required_pops = {x for x, y in self.pops.items() if y["type"] == tdve.pop_type}  # The TDVE should contain values for all populations of that type, otherwise cannot construct the ParameterSet. Check that these populations are all present

                        # Check that data is available for all populations. If the TDVE table contains an entry for 'all'
                        # then further checks are not necessary because a fallback value will be available for every population
                        tdve_pops = tdve.ts.keys()
                        if not set(tdve_pops).intersection({"all", "All"}):
                            missing_pops = required_pops.difference(tdve_pops)
                            if missing_pops:
                                raise InvalidDatabook("%s. The following populations were not supplied but are required: %s" % (location, missing_pops))

                        for name, ts in self.tdve[spec_name].ts.items():
                            assert ts.has_data, "%s. Data values missing for %s (%s)" % (location, tdve.name, name)
                            assert ts.units is not None, "%s. Units missing for %s (%s)" % (location, tdve.name, name)
                            if ts.units.strip().lower() != framework_units.strip().lower():
                                # If the units don't match the framework's 'databook' units, see if they at least match the standard unit (for legacy databooks)
                                # For compartments and characteristics, the units must match exactly
                                if obj_type in ["comps", "characs"] or ("format" in spec and spec["format"] is not None and ts.units.lower().strip() != spec["format"].lower().strip()):
                                    assert ts.units == framework_units, '%s. Unit "%s" for %s (%s) does not match the declared units from the Framework (expecting "%s")' % (location, ts.units, tdve.name, name, framework_units)
                            if obj_type == "par" and spec["timed"] == "y":
                                assert not ts.has_time_data, "%s. Parameter %s (%s) is marked as a timed transition in the Framework, so it must have a constant value (i.e., the databook cannot contain time-dependent values for this parameter)" % (location, tdve.name, name)

        for tdc in self.interpops + self.transfers:
            if tdc.from_pop_type is None:  # Supply default pop type
                tdc.from_pop_type = self._pop_types[0]
            assert tdc.from_pop_type in self._pop_types, 'Error in transfer/interaction "%s": from population type "%s" not found in framework. If the framework defines a non-default population type, then it must be explicitly specified in databooks and program books.' % (tdc.full_name, tdc.from_pop_type)
            if tdc.to_pop_type is None:  # Supply default pop type
                tdc.to_pop_type = self._pop_types[0]
            assert tdc.to_pop_type in self._pop_types, 'Error in transfer/interaction "%s": to population type "%s" not found in framework. If the framework defines a non-default population type, then it must be explicitly specified in databooks and program books.' % (tdc.full_name, tdc.to_pop_type)

        for _, spec in framework.interactions.iterrows():
            for tdc in self.interpops:
                if tdc.code_name == spec.name:
                    for (from_pop, to_pop), ts in tdc.ts.items():
                        assert to_pop in self.pops, 'Population "%s" in "%s" not recognized. Should be one of: %s' % (to_pop, spec.name, self.pops.keys())
                        assert self.pops[to_pop]["type"] == tdc.to_pop_type, 'Interaction "%s" has to-population type "%s", but contains Population "%s", which is type "%s"' % (tdc.full_name, tdc.to_pop_type, to_pop, self.pops[to_pop]["type"])
                        assert from_pop in self.pops, 'Population "%s" in "%s" not recognized. Should be one of: %s' % (from_pop, spec.name, self.pops.keys())
                        assert self.pops[from_pop]["type"] == tdc.from_pop_type, 'Interaction "%s" has from-population type "%s", but contains Population "%s", which is type "%s"' % (tdc.full_name, tdc.from_pop_type, from_pop, self.pops[from_pop]["type"])
                        assert ts.has_data, "Data values missing for interaction %s, %s->%s" % (spec.name, to_pop, from_pop)
                        assert ts.units.lower().title() == FS.DEFAULT_SYMBOL_INAPPLICABLE.lower().title(), 'Units error in interaction %s, %s->%s. Interaction units must be "N.A."' % (spec.name, to_pop, from_pop)
                    break
            else:
                raise InvalidDatabook('Required interaction "%s" not found in databook' % spec.name)

        for tdc in self.transfers:
            for (from_pop, to_pop), ts in tdc.ts.items():
                assert to_pop in self.pops, 'Population "%s" in "%s" not recognized. Should be one of: %s' % (to_pop, tdc.full.name, self.pops.keys())
                assert self.pops[to_pop]["type"] == tdc.to_pop_type, 'Transfer "%s" has population type "%s", but contains Population "%s", which is type "%s"' % (tdc.full_name, tdc.to_pop_type, to_pop, self.pops[to_pop]["type"])
                assert from_pop in self.pops, 'Population "%s" in "%s" not recognized. Should be one of: %s' % (from_pop, tdc.full.name, self.pops.keys())
                assert self.pops[from_pop]["type"] == tdc.from_pop_type, 'Transfer "%s" has population type "%s", but contains Population "%s", which is type "%s"' % (tdc.full_name, tdc.from_pop_type, from_pop, self.pops[from_pop]["type"])
                assert ts.has_data, "Data values missing for transfer %s, %s->%s" % (tdc.full_name, to_pop, from_pop)
                assert ts.units is not None, "Units are missing for transfer %s, %s->%s" % (tdc.full_name, to_pop, from_pop)
        return True

    def to_workbook(self) -> tuple:
        """
        Return an open workbook for the databook

        This allows the xlsxwriter workbook to be manipulated prior to closing the
        filestream e.g. to append extra sheets. This prevents issues related to cached
        data values when reloading a workbook to append or modify content

        Warning - the workbook is backed by a BytesIO instance and needs to be closed.
        See the usage of this method in the :meth`to_spreadsheet` function.

        :return: A tuple (bytes, workbook) with a BytesIO instance and a corresponding *open* xlsxwriter workbook instance

        """

        # Initialize the bytestream
        f = io.BytesIO()
        wb = xw.Workbook(f, {"in_memory": True})

        # Open a workbook
        self._book = wb
        self._book.set_properties({"category": "atomica:databook"})
        self._formats = standard_formats(self._book)
        self._references = {}  # Reset the references dict

        # Write the contents
        self._write_pops()
        self._write_tdve()
        self._write_interpops()
        self._write_transfers()

        # Clean internal variables related to writing the worbkook
        self._book = None
        self._formats = None
        self._references = None

        return f, wb

    def to_spreadsheet(self) -> sc.Spreadsheet:
        """
        Return content as a Sciris Spreadsheet

        :return: A :class:`sciris.Spreadsheet` instance

        """

        f, wb = self.to_workbook()
        wb.close()  # Close the workbook to flush any xlsxwriter content
        spreadsheet = sc.Spreadsheet(f)  # Wrap it in a spreadsheet instance
        return spreadsheet

    def save(self, fname) -> None:
        """
        Save databook to disk

        This function provides a shortcut to generate a spreadsheet and immediately save it to disk.

        :param fname: File name to write on disk

        """

        ss = self.to_spreadsheet()
        ss.save(fname)

    def add_pop(self, code_name: str, full_name: str, pop_type: str = None) -> None:
        """
        Add a population

        This will add a population to the databook. The population type should match
        one of the population types in the framework

        :param code_name: The code name for the new population
        :param full_name: The full name/label for the new population
        :param pop_type: String with the population type code name (optional) - default is the type of the first population

        """

        if pop_type is None:
            pop_type = self._pop_types[0]
        assert pop_type in self._pop_types, 'Population type "%s" not found in framework' % (pop_type)

        code_name = code_name.strip()
        assert len(code_name) > 1, 'Population code name (abbreviation) "%s" is not valid - it must be at least two characters long' % (code_name)
        assert code_name not in self.pops, 'Population with name "%s" already exists' % (code_name)

        if code_name.lower() in FS.RESERVED_KEYWORDS:
            raise Exception('Population name "%s" is a reserved keyword' % (code_name.lower()))

        self.pops[code_name] = {"label": full_name, "type": pop_type}

        for interaction in self.transfers + self.interpops:
            if interaction.from_pop_type == pop_type:
                interaction.from_pops.append(code_name)
            if interaction.to_pop_type == pop_type:
                interaction.to_pops.append(code_name)

        for tdve in self.tdve.values():
            # Since TDVEs in databooks must have the unit set in the framework, all ts objects must share the same units
            # And, there is only supposed to be one type of unit allowed for TDVE tables (if the unit is empty, it will be 'N.A.')
            # so can just pick the first of the allowed units. We will add the population row if the pop type matches and if
            # the TDVE is either not a 'default_all' or if the user has removed the 'All' row from the TDVE despite it being default_all
            if tdve.pop_type != pop_type or (tdve.default_all and ("All" in tdve.ts or "all" in tdve.ts)):
                continue
            else:
                tdve.ts[code_name] = TimeSeries(units=tdve.allowed_units[0])

    def rename_pop(self, existing_code_name: str, new_code_name: str, new_full_name: str) -> None:
        """
        Rename a population

        :param existing_code_name: Existing code name of a population
        :param new_code_name: New code name to assign
        :param new_full_name: New full name/label to assign

        """

        existing_code_name = existing_code_name.strip()
        new_code_name = new_code_name.strip()
        assert len(new_code_name) > 1, 'New population code name (abbreviation) "%s" is not valid - it must be at least two characters long' % (new_code_name)

        assert existing_code_name in self.pops, 'A population with code name "%s" is not present' % (existing_code_name)
        assert new_code_name not in self.pops, 'Population with name "%s" already exists' % (new_code_name)

        if new_code_name.lower() in FS.RESERVED_KEYWORDS:
            raise Exception('Population name "%s" is a reserved keyword' % (new_code_name.lower()))

        # First change the name of the key
        self.pops.rename(existing_code_name, new_code_name)

        # Then change the full name
        self.pops[new_code_name]["label"] = new_full_name

        # Update interactions and transfers - need to change all of the to/from tuples
        for interaction in self.transfers + self.interpops:
            idx = interaction.from_pops.index(existing_code_name)
            interaction.from_pops[idx] = new_code_name

            idx = interaction.to_pops.index(existing_code_name)
            interaction.to_pops[idx] = new_code_name

            for from_pop, to_pop in interaction.ts.keys():
                if to_pop == existing_code_name and from_pop == existing_code_name:
                    interaction.ts.rename((from_pop, to_pop), (new_code_name, new_code_name))
                elif from_pop == existing_code_name:
                    interaction.ts.rename((from_pop, to_pop), (new_code_name, to_pop))
                elif to_pop == existing_code_name:
                    interaction.ts.rename((from_pop, to_pop), (from_pop, new_code_name))

        # Update TDVE tables
        for tdve in self.tdve.values():
            if existing_code_name in tdve.ts:
                tdve.ts.rename(existing_code_name, new_code_name)

    def remove_pop(self, pop_name):
        # Remove population with given code name
        del self.pops[pop_name]

        for interaction in self.transfers + self.interpops:
            interaction.to_pops.remove(pop_name)
            interaction.from_pops.remove(pop_name)

            for k in list(interaction.ts.keys()):
                if k[0] == pop_name or k[1] == pop_name:
                    del interaction.ts[k]

        for tdve in self.tdve.values():
            for k in list(tdve.ts.keys()):
                if k == pop_name:
                    del tdve.ts[k]

    def _assert_unique_tdc(self, name: str) -> None:
        """
        Check if a TDC name already exists

        Used when adding or renaming transfers and interactions

        :param name: Proposed name of new TDC
        :return: None
        :raises: Exception if the name is already present
        """
        for tdc in self.transfers + self.interpops:
            if name == tdc.code_name:
                raise Exception(f'The databook already contains an existing {tdc.type} with name "{name}"')

    def add_transfer(self, code_name: str, full_name: str, pop_type: str = None) -> TimeDependentConnections:
        """
        Add a new empty transfer

        :param code_name: The code name of the transfer to create
        :param full_name: The full name of the transfer to create
        :param pop_type: Code name of the population type. Default is first population type in the framework
        :return: Newly instantiated TimeDependentConnections object (also added to ``ProjectData.transfers``)

        """

        if pop_type is None:
            pop_type = self._pop_types[0]

        assert pop_type in self._pop_types, "Population type %s not found in framework" % (pop_type)
        self._assert_unique_tdc(code_name)

        pop_names = [name for name, pop_spec in self.pops.items() if pop_spec["type"] == pop_type]

        # Here, need to list all relevant populations
        new_transfer = TimeDependentConnections(code_name, full_name, self.tvec, from_pops=pop_names, to_pops=pop_names, interpop_type="transfer", ts=None, from_pop_type=pop_type, to_pop_type=pop_type)
        new_transfer.write_units = True
        new_transfer.write_assumption = True
        new_transfer.write_uncertainty = True
        self.transfers.append(new_transfer)
        return new_transfer

    def rename_transfer(self, existing_code_name: str, new_code_name: str, new_full_name: str) -> None:
        """
        Rename an existing transfer

        :param existing_code_name: The existing code name to change
        :param new_code_name: The new code name
        :param new_full_name: The new full name

        """

        # Find the transfer to change
        for transfer in self.transfers:
            if existing_code_name == transfer.code_name:
                transfer_to_change = transfer
                break
        else:
            raise NotFoundError('Transfer with name "%s" was not found' % (existing_code_name))

        # Check no name collisions
        self._assert_unique_tdc(new_code_name)

        # Modify it
        transfer_to_change.code_name = new_code_name
        transfer_to_change.full_name = new_full_name

    def remove_transfer(self, code_name: str) -> None:
        """
        Remove a transfer

        :param code_name: Code name of the transfer to remove
        """

        names = [x.code_name for x in self.transfers]
        idx = names.index(code_name)
        del self.transfers[idx]

    # NB. Differences in the model will only happen if the model knows what to do with the new interaction
    def add_interaction(self, code_name: str, full_name: str, from_pop_type: str = None, to_pop_type: str = None) -> TimeDependentConnections:
        """
        Add a new empty interaction

        Normally this method would only be manually called if a framework had been
        updated to contain a new interaction, and the databook now required updating.
        Therefore, this method would generally only be used when an interaction
        with given code name, full name, and pop type had already been added to a framework.

        :param code_name: The code name of the interaction to create
        :param full_name: The full name of the interaction to create
        :param from_pop_type: The name of a population type, which will identify the populations to be added. Default is first population type in the framework
        :param to_pop_type: The name of a population type, which will identify the populations to be added. Default is first population type in the framework
        :return: Newly instantiated TimeDependentConnections object (also added to ``ProjectData.interpops``)

        """

        if from_pop_type is None:
            from_pop_type = self._pop_types[0]
        if to_pop_type is None:
            to_pop_type = self._pop_types[0]

        assert from_pop_type in self._pop_types, "Population type %s not found in framework" % (from_pop_type)
        assert to_pop_type in self._pop_types, "Population type %s not found in framework" % (to_pop_type)

        self._assert_unique_tdc(code_name)

        from_pops = [name for name, pop_spec in self.pops.items() if pop_spec["type"] == from_pop_type]
        to_pops = [name for name, pop_spec in self.pops.items() if pop_spec["type"] == to_pop_type]
        interpop = TimeDependentConnections(code_name, full_name, tvec=self.tvec, from_pops=from_pops, to_pops=to_pops, interpop_type="interaction", ts=None, from_pop_type=from_pop_type, to_pop_type=to_pop_type)
        interpop.write_units = True
        interpop.write_assumption = True
        interpop.write_uncertainty = True
        self.interpops.append(interpop)
        return interpop

    def remove_interaction(self, code_name: str) -> None:
        """
        Remove an interaction

        :param code_name: Code name of the interaction to remove
        """

        names = [x.code_name for x in self.interpops]
        idx = names.index(code_name)
        del self.interpops[idx]

    def _read_pops(self, sheet) -> None:
        """
        Reads the 'Population Definitions' sheet

        """

        # TODO - can modify _read_pops() and _write_pops() if there are more population attributes
        tables = read_tables(sheet)[0]
        assert len(tables) == 1, "Population Definitions page should only contain one table"

        self.pops = sc.odict()
        assert cell_get_string(tables[0][0][0]).lower() == "abbreviation"
        assert cell_get_string(tables[0][0][1]).lower() == "full name"

        # If pop typ column exists, check the heading is correct
        if len(tables[0][0]) > 2:
            assert cell_get_string(tables[0][0][2]).lower() == "population type"

        for row in tables[0][1:]:

            pop_name = cell_get_string(row[0])
            assert len(pop_name) > 1, 'Population code name (abbreviation) "%s" is not valid - it must be at least two characters long' % (pop_name)

            label = cell_get_string(row[1])
            assert len(label) > 1, 'Population full name "%s" is not valid - it must be at least two characters long' % (label)

            if pop_name.lower() in FS.RESERVED_KEYWORDS:
                raise Exception('Population name "%s" is a reserved keyword' % (pop_name.lower()))

            poptype = None
            if len(row) > 2 and row[2].value is not None:
                poptype = cell_get_string(row[2])

            self.pops[pop_name] = {"label": label, "type": poptype}

    def _write_pops(self) -> None:
        """
        Writes the 'Population Definitions' sheet

        """

        sheet = self._book.add_worksheet("Population Definitions")
        sheet.set_tab_color("#FFC000")  # this tab is orange
        widths = dict()

        current_row = 0
        sheet.write(current_row, 0, "Abbreviation", self._formats["center_bold"])
        update_widths(widths, 0, "Abbreviation")
        sheet.write(current_row, 1, "Full Name", self._formats["center_bold"])
        update_widths(widths, 1, "Abbreviation")
        sheet.write(current_row, 2, "Population type", self._formats["center_bold"])
        update_widths(widths, 2, "Population type")

        for name, content in self.pops.items():
            current_row += 1
            sheet.write(current_row, 0, name, self._formats["unlocked"])
            update_widths(widths, 0, name)
            sheet.write(current_row, 1, content["label"], self._formats["unlocked"])
            update_widths(widths, 1, content["label"])
            sheet.write(current_row, 2, content["type"], self._formats["not_required"])
            update_widths(widths, 2, content["type"])

            self._references[name] = "='%s'!%s" % (sheet.name, xlrc(current_row, 0, True, True))
            self._references[content["label"]] = "='%s'!%s" % (sheet.name, xlrc(current_row, 1, True, True))  # Reference to the full name

        apply_widths(sheet, widths)

    def _read_transfers(self, sheet) -> None:
        """
        Reads the 'Transfers' sheet

        """
        tables, start_rows = read_tables(sheet)
        assert len(tables) % 3 == 0, "There should be 3 subtables for every transfer"
        self.transfers = []
        names = set()
        for i in range(0, len(tables), 3):
            transfer = TimeDependentConnections.from_tables(tables[i : i + 3], "transfer")
            if transfer.code_name in names:
                raise Exception(f"Another transfer with name '{transfer.code_name}' already exists")
            names.add(transfer.code_name)
            self.transfers.append(transfer)

    def _write_transfers(self) -> None:
        """
        Writes the 'Transfers' sheet

        """
        # Writes a sheet for every transfer

        # Skip if no transfers
        if not self.transfers:
            return

        sheet = self._book.add_worksheet("Transfers")
        sheet.set_tab_color("#808080")
        # sheet.hide()
        widths = dict()
        next_row = 0
        for transfer in self.transfers:
            next_row = transfer.write(sheet, next_row, self._formats, self._references, widths)
        apply_widths(sheet, widths)

    def _read_interpops(self, sheet) -> None:
        """
        Reads the 'Interactions' sheet

        """
        tables, start_rows = read_tables(sheet)
        assert len(tables) % 3 == 0, "There should be 3 subtables for every interaction"
        self.interpops = []
        names = set()
        for i in range(0, len(tables), 3):
            interaction = TimeDependentConnections.from_tables(tables[i : i + 3], "interaction")
            if interaction.code_name in names:
                raise Exception(f"Another interaction with name '{interaction.code_name}' already exists")
            names.add(interaction.code_name)
            self.interpops.append(interaction)

        return

    def _write_interpops(self) -> None:
        """
        Writes the 'Interactions' sheet

        """

        # Skip if no interpops
        if not self.interpops:
            return

        sheet = self._book.add_worksheet("Interactions")
        sheet.set_tab_color("#808080")
        widths = dict()
        next_row = 0
        for interpop in self.interpops:
            next_row = interpop.write(sheet, next_row, self._formats, self._references, widths)
        apply_widths(sheet, widths)

    def _write_tdve(self) -> None:
        """
        Writes the TDVE tables

        This method will create multiple sheets, one for each custom page specified
        in the Framework.

        """

        for sheet_name, code_names in self.tdve_pages.items():
            sheet = self._book.add_worksheet(sheet_name)
            widths = dict()

            next_row = 0
            has_editable_content = False
            for code_name in code_names:
                has_editable_content = has_editable_content or (not self.tdve[code_name].has_data)  # there is editable content if any TDVE is missing data, so blue cells are present
                next_row = self.tdve[code_name].write(sheet, next_row, self._formats, references=self._references, widths=widths)

            if has_editable_content:
                sheet.set_tab_color("#92D050")
            else:
                sheet.set_tab_color("#808080")

            apply_widths(self._book.get_worksheet_by_name(sheet_name), widths)



"""
Manage Project versions and migration

Migration refers to updating old :class:`Project` instances so that they
can be run with more recent versions of Atomica. This module defines

- A set of 'migration functions' that each transform a :class:`Project` instance from one
  version to another
- An entry-point that sequentially calls the migration functions to update a project to
  the current version used in Atomica

"""

import sys
import io
from packaging.version import Version
from .system import logger
from .version import version, gitinfo
import sciris as sc
from .system import FrameworkSettings as FS
import atomica
import types
import numpy as np
import pandas as pd
from collections import defaultdict
import pathlib
import os

__all__ = ["migration", "migrate", "register_migration"]

SKIP_MIGRATION = False  # Global migration flag to disable migration

# MODULE MIGRATIONS
#
# In this section, make changes to the Atomica module structure to enable pickle files
# to be loaded at all

# On Windows, pathlib.PosixPath is not implemented which can cause problems when unpickling
# Therefore, we need to turn PosixPath into WindowsPath on loading. The reverse is true on
# Unix systems
if os.name == "nt":
    pathlib.PosixPath = pathlib.WindowsPath
else:
    pathlib.WindowsPath = pathlib.PosixPath


class _Placeholder:
    pass


# First, add any placeholder modules that have been subsequently removed
atomica.structure = types.ModuleType("structure")  # Removed 'structure' module in 1.0.12 20181107
sys.modules["atomica.structure"] = atomica.structure

# Then, remap any required classes
atomica.structure.TimeSeries = atomica.utils.TimeSeries  # Moved 'TimeSeries' in 1.0.12 20181107

# If classes have been removed, then use the Placeholder class to load them
# The migration function can then replace Placeholder instances with actual
# instances - see the AtomicaSpreadsheet -> sciris.Spreadsheet migration function
atomica.excel.AtomicaSpreadsheet = _Placeholder
atomica.optimization.OptimInstructions = _Placeholder

# PROJECT MIGRATIONS
#
# The remaining code manages upgrading Project objects and their contents after they
# have been unpickled and instantiated

# This dict stores the migrations associated with each versioned class
# We migrate projects and results separately because they might be stored
# separately e.g. in a database


class Migration:
    """Class representation of a migration

    This class stores a migration function together with all required metadata. It would
    normally be instantiated using the `migration` decorator, which also registers the
    migration by adding it to the `migrations` list

    """

    def __init__(self, classname, original_version, new_version, description, fcn, date=None, update_required=False):
        """

        :param original_version:
        :param new_version:
        :param description:
        :param fcn:
        :param date:
        :param update_required: Flag if dependent content in the object might change e.g. if redoing a calculation might produce different results
        """
        self.classname = classname
        self.original_version = original_version
        self.new_version = new_version
        self.description = description
        self.update_required = update_required
        self.date = date
        self.fcn = fcn

    def upgrade(self, obj):
        logger.debug("MIGRATION: Upgrading %s %s -> %s (%s)" % (self.classname, self.original_version, self.new_version, self.description))
        obj = self.fcn(obj)  # Run the migration function
        if obj is None:
            raise Exception("%s returned None, it is likely missing a return statement" % (str(self)))
        obj.version = self.new_version  # Update the version
        if self.update_required:
            obj._update_required = True
        return obj

    def __repr__(self):
        return f"Migration({self.classname}, {self.original_version}->{self.new_version})"


def register_migration(registry, classname, original_version, new_version, description, date=None, update_required=False):
    """Decorator to register migration functions

    This decorator constructs a `Migration` object from a decorated migration function, and registers it
    in the module's list of migrations to be run when calling migrate()

    :param registry: Dictionary storing {classname:[migrations]}
    :param original_version: Version string for the start version. Function will only run on projects whose version is <= this
    :param new_version: The resulting project is assigned this version
    :param description: A brief overview of the purpose of this migration
    :param date: Optionally specify the date when this migration was written
    :param update_required: Optionally flag that changes to model output may occur - the flag is stored in the object's `_update_required` attribute.
                            Objects can optionally contain this attribute, so it's not necessary if the object is planned not to require
                            such a flag.
    :return: None

    Example usage::

        @migration('Project','1.0.0', '1.0.1', 'Upgrade project', update_required=True)
        def _update_project(proj):
            ...
            return proj

    The migration function (update_project()) takes a single argument which is a project object, and returns
    a project object. Decorating the function registers the migration and it will automatically be used by
    migrate() and therefore Project.load()

    """

    def register(f):
        if classname not in registry:
            registry[classname] = []
        registry[classname].append(Migration(classname, original_version, new_version, description, fcn=f, date=date, update_required=update_required))
        return f

    return register


migrations = dict()  # Registry of migrations in Atomica


def migration(*args, **kwargs):
    # Wrapper decorator to bind register_migration to the migration registry in this module
    return register_migration(migrations, *args, **kwargs)


def migrate(obj, registry=migrations, version=version, gitinfo=gitinfo):
    """Update a object to the latest version

    Run all of the migrations in the list of available migrations. The migrations are run in ascending order, as long
    as the version is <= the migration's original version. This way, migrations don't need to be added if a version number
    change takes place without actually needing a migration - instead, when adding a migration, just use whatever version
    numbers are appropriate at the time the change is introduced, it will behave sensibly.

    Typically, this function would not be called manually - it happens automatically as part of `Project.load()`.

    Note that this function returns the updated object. Typically the migration functions will update a
    object in-place, but this syntax allows migrations to replace parts of the object (or the entire object)
    if that is ever necessary in the future. So do not rely on `migrate` having side-effects, make sure the
    returned object is retained. In principle we could dcp() the object to ensure that only the
    copy gets migrated, but at this point it does not appear worth the performance overhead (since migration at the
    moment is only ever called automatically).

    :param proj: object to migrate
    :param version: The current version (in case this is reused in a separate module, defaults to Atomica's)
    :param gitinfo: The gitinfo to update (in case this is reused in a separate module, defaults to Atomica's)
    :return: Updated object

    """

    if type(obj).__name__ not in registry:
        return obj  # If there are no migrations for the object, then return immediately

    elif not hasattr(obj, "version"):
        # If the object has no version attribute, then add one with version 0. This is presumably
        # because the original object didn't have a version, but now migrations for it are required.
        # In that case, any object that doesn't have a version would require all migrations to be run.
        # New objects would all have a version already. After any migrations are run, these values will
        # then be updated to the current version
        obj.version = "0.0.0"
        obj.gitinfo = None

    if SKIP_MIGRATION:
        print("Skipping migration")
        return obj  # If migration is disabled then don't make any changes EXCEPT to add in version and gitinfo which may otherwise be hard to catch

    migrations_to_run = sorted(registry[type(obj).__name__], key=lambda m: Version(m.original_version))
    if sc.compareversions(obj.version, version) >= 0:
        return obj
    else:
        if hasattr(obj, "name"):
            logger.info('Migrating %s "%s" from %s->%s', type(obj).__name__, obj.name, obj.version, version)
        else:
            logger.info("Migrating %s from %s->%s", type(obj).__name__, obj.version, version)

    for m in migrations_to_run:  # Run the migrations in increasing version order
        if sc.compareversions(obj.version, m.new_version) < 0:
            obj = m.upgrade(obj)

    obj.version = version  # Set object version to the current Atomica version
    obj.gitinfo = gitinfo  # Update gitinfo to current version
    if hasattr(obj, "_update_required") and obj._update_required:
        logger.warning("Caution: due to migration, object may behave different if re-run.")
    return obj


def all_results(proj):
    """Helper generator to iterate over all results in a project

    Project results may be
        - Standalone results
        - A list of results
        - A UID string (that should be ignored)

    This function is a generator that allows iteration over all results in a project

    :param proj: A Project object
    :return: A Result (via yield)

    Example usage:

    for result in all_results(proj):
        do_stuff(result)

    """
    from .results import Result

    for result in proj.results.values():
        if isinstance(result, list):
            for r in result:
                yield r
        elif isinstance(result, Result):
            yield result


def all_progsets(proj):
    """Helper generator to iterate over all progsets in a project

    Project progsets may be
        - Standalone progsets
        - Contained in a Result

    This function is a generator that allows iteration over all progsets in a project

    :param proj: A Project object
    :return: A progset (via yield)

    Example usage:

    for progset in all_progsets(proj):
        do_stuff(progset)

    """
    from .results import Result

    for progset in proj.progsets.values():
        yield (progset)

    for result in proj.results.values():
        if isinstance(result, list):
            for r in result:
                if r.model.progset is not None:
                    yield r.model.progset
        elif isinstance(result, Result) and result.model.progset is not None:
            yield result.model.progset


def all_frameworks(proj):
    """Helper generator to iterate over all frameworks in a project

    Project frameworks may be
        - Standalone frameworks
        - Contained in a Result

    This function is a generator that allows iteration over all frameworks in a project

    :param proj: A Project object
    :return: A framework (via yield)

    """
    from .results import Result

    if proj.framework:
        yield proj.framework

    for result in proj.results.values():
        if isinstance(result, list):
            for r in result:
                yield r.framework
        elif isinstance(result, Result):
            yield result.framework


@migration("Project", "1.0.5", "1.0.6", "Simplify ParameterSet storage")
def _simplify_parset_storage(proj):
    # ParameterSets in 1.0.5 store the parameters keyed by the type e.g. 'cascade','comp'
    # In 1.0.6 they are flat
    for parset in proj.parsets.values():
        new_pars = sc.odict()
        for par_type, par_list in parset.pars.items():
            for par in par_list:
                new_pars[par.name] = par
        parset.pars = new_pars
        del parset.par_ids
    return proj


@migration("Project", "1.0.7", "1.0.8", "Add version information to model")
def _add_model_version(proj):
    # Note that this is a legacy update via the Project and therefore Results that
    # are stored separately will not receive this migration. It is unlikely that any
    # Results have been stored in this form though.
    def add_version(res, p):
        res.model.version = p.version
        res.model.gitinfo = p.gitinfo
        res.model.created = p.created

    for result in all_results(proj):
        add_version(result, proj)

    return proj


@migration("ProgramSet", "1.0.8", "1.0.9", "Add currency and units to progset quantities")
def _add_currency_and_units(progset):

    # If the progset was saved prior to 1.23, it would have no version but it might have already
    # had these migrations applied. Therefore we need to check if they have already been run
    if hasattr(progset, "currency"):
        return progset

    progset.currency = "$"
    for program in progset.programs.values():
        program.baseline_spend.units = progset.currency + "/year"
        program.spend_data.units = progset.currency + "/year"
        program.unit_cost.units = progset.currency + "/person"
        program.capacity.units = "people/year"
        program.saturation.units = FS.DEFAULT_SYMBOL_INAPPLICABLE
        program.coverage.units = "people/year"

    return progset


@migration("ProgramSet", "1.0.9", "1.0.10", "Remove target_pars from Programs")
def _remove_target_pars(progset):

    for program in progset.programs.values():
        if hasattr(program, "target_pars"):
            del program.target_pars

    return progset


@migration("Project", "1.0.10", "1.0.11", "Add result update flag to Project")
def _add_result_update_flag(proj):
    proj._update_required = False
    return proj


@migration("Project", "1.0.12", "1.0.13", "Add timescale to parameters")
def _add_timescale(proj):
    for _, spec in proj.framework.pars.iterrows():
        if proj.framework.transitions[spec.name]:
            proj.framework.pars.at[spec.name, "timescale"] = 1.0  # Default timescale - note that currently only transition parameters are allowed to have a timescale that is not None

            if not spec["format"]:
                if not proj.data:
                    raise Exception('The Framework now needs to define the units for transition parameters, but the current one does not define units for Parameter "%s" and data was not saved either, so cannot infer the unit type. Please modify the Framework spreadsheet to add units' % (spec.name))
                else:
                    # Get the units from the first ts associated with the table
                    units = set([x.units.lower().strip() for x in proj.data.tdve[spec.name].ts.values()])
                    if len(units) == 1:
                        proj.framework.pars.at[spec.name, "format"] = list(units)[0]
                    else:
                        raise Exception('Parameters must now have a single unit for all populations. However, the existing data has more than one unit type associated with Parameter "%s" so it is no longer valid.' % (spec.name))

    return proj


@migration("Project", "1.0.13", "1.0.14", "Parameters use TimeSeries internally")
def _parameter_use_timeseries(proj):
    for parset in proj.parsets.values():
        for par in parset.all_pars():

            par.ts = dict()
            for pop in par.t.keys():
                par.ts[pop] = atomica.TimeSeries(units=par.y_format[pop])
                if par.t[pop] is None:
                    continue
                elif len(par.t[pop]) == 1 and not np.isfinite(par.t[pop][0]):
                    # It was an assumption
                    par.ts[pop].assumption = par.y[pop][0]
                else:
                    par.ts[pop].t = par.t[pop].tolist()
                    par.ts[pop].vals = par.y[pop].tolist()

            del par.y_format
            del par.t
            del par.y
            del par.autocalibrate

    return proj


@migration("Project", "1.0.14", "1.0.15", "Internal model tidying")
def _model_tidying(proj):
    for result in all_results(proj):
        for pop in result.model.pops:
            for charac in pop.characs:
                charac._vals = charac.internal_vals
                charac._is_dynamic = charac.dependency
                del charac.internal_vals
                del charac.dependency
            for par in pop.pars:
                par._is_dynamic = par.dependency
                par._precompute = False
                del par.dependency
    return proj


def _convert_atomica_spreadsheet(placeholder):
    new = sc.Spreadsheet(source=io.BytesIO(placeholder.data), filename=placeholder.filename)
    new.created = placeholder.load_date
    new.modified = placeholder.load_date
    return new


@migration("Project", "1.0.15", "1.0.16", "Replace AtomicaSpreadsheet in project")
def _convert_project_spreadsheets(proj):
    if proj.databook is not None:
        proj.databook = _convert_atomica_spreadsheet(proj.databook)
    if proj.progbook is not None:
        proj.progbook = _convert_atomica_spreadsheet(proj.progbook)
    return proj


@migration("ProjectFramework", "1.0.15", "1.0.16", "Replace AtomicaSpreadsheet in framework")
def _convert_framework_spreadsheets(framework):
    if framework.spreadsheet is not None:
        framework.spreadsheet = _convert_atomica_spreadsheet(framework.spreadsheet)
    return framework


@migration("ProgramSet", "1.0.16", "1.0.17", "Rename capacity constraint")
def _rename_capacity_constraint(progset):
    for prog in progset.programs.values():
        if hasattr(prog, "capacity"):
            prog.capacity_constraint = prog.capacity
            del prog.capacity
    return progset


@migration("Result", "1.0.27", "1.0.28", "Rename link labels")
def _rename_link_labels(result):
    for pop in result.model.pops:
        for link in pop.links:
            link.id = link.id[0:3] + (link.parameter.name + ":flow",)
    result.model._set_vars_by_pop()
    return result


@migration("Project", "1.0.30", "1.1.3", "Replace scenarios")
def _replace_scenarios(proj):
    # This migration upgrades existing scenarios to match the latest definitions
    from .scenarios import ParameterScenario, BudgetScenario, CoverageScenario
    from .utils import NDict, TimeSeries

    new_scens = NDict()

    for name, scen in proj.scens.items():

        scen_name = scen.name
        active = scen.active

        try:
            parsetname = proj.parset(scen.parsetname).name
        except Exception:
            parsetname = proj.parsets[-1].name

        if isinstance(scen, ParameterScenario):
            new_scen = scen  # No need to migrate parameter scenarios

        elif isinstance(scen, BudgetScenario):
            # Convert budget scenario to instructions based on existing logic
            if scen.alloc_year is not None:
                # If the alloc_year is prior to the program start year, then just use the spending value directly for all times
                # For more sophisticated behaviour, the alloc should be passed into the BudgetScenario as a TimeSeries
                alloc = sc.odict()
                for prog_name, val in scen.alloc.items():
                    assert not isinstance(val, TimeSeries)  # Value must not be a TimeSeries
                    if val is None:
                        continue  # Use default spending for any program that does not have a spending overwrite
                    alloc[prog_name] = TimeSeries(scen.alloc_year, val)
                    if scen.alloc_year > scen.start_year and proj.progsets:
                        # Add in current spending if a programs instance already exists
                        progset = proj.progsets[-1]
                        # If adding spending in a future year, linearly ramp from the start year
                        spend_data = progset.programs[prog_name].spend_data
                        alloc[prog_name].insert(scen.start_year, spend_data.interpolate(scen.start_year))  # This will result in a linear ramp
            else:
                alloc = sc.odict()
                for prog_name, val in scen.alloc.items():
                    if not isinstance(val, TimeSeries):
                        alloc[prog_name] = TimeSeries(scen.start_year, val)
                    else:
                        alloc[prog_name] = sc.dcp(val)
            for ts in alloc.values():
                ts.vals = [x * scen.budget_factor for x in ts.vals]

            try:
                progsetname = proj.progset(scen.progsetname).name
            except Exception:
                progsetname = proj.parsets[-1].name

            new_scen = atomica.BudgetScenario(name=scen_name, active=active, parsetname=parsetname, progsetname=progsetname, alloc=alloc, start_year=scen.start_year)

        elif isinstance(scen, CoverageScenario):
            coverage = sc.odict()
            for prog_name, val in scen.coverage.items():
                if not isinstance(val, TimeSeries):
                    coverage[prog_name] = TimeSeries(scen.start_year, val)
                else:
                    coverage[prog_name] = sc.dcp(val)

            try:
                progsetname = proj.progset(scen.progsetname).name
            except Exception:
                progsetname = proj.parsets[-1].name

            new_scen = atomica.CoverageScenario(name=scen_name, active=active, parsetname=parsetname, progsetname=progsetname, coverage=coverage, start_year=scen.start_year)

        new_scens.append(new_scen)

    proj.scens = new_scens
    return proj


@migration("ProgramSet", "1.2.0", "1.3.0", "Add population type")
def _add_pop_type_progset(progset):
    for pop in progset.pops.keys():
        if sc.isstring(progset.pops[pop]):
            progset.pops[pop] = {"label": progset.pops[pop], "type": FS.DEFAULT_POP_TYPE}

    for comp in progset.comps.keys():
        if sc.isstring(progset.comps[comp]):
            progset.comps[comp] = {"label": progset.comps[comp], "type": FS.DEFAULT_POP_TYPE}

    for par in progset.pars.keys():
        if sc.isstring(progset.pars[par]):
            progset.pars[par] = {"label": progset.pars[par], "type": FS.DEFAULT_POP_TYPE}

    return progset


@migration("Project", "1.2.0", "1.3.0", "Add population type")
def _add_pop_type(proj):

    for fw in all_frameworks(proj):

        # Add default population type sheet
        fw.sheets["population types"] = [pd.DataFrame.from_records([(FS.DEFAULT_POP_TYPE, "Default")], columns=["code name", "description"])]
        fw.comps["population type"] = FS.DEFAULT_POP_TYPE
        fw.characs["population type"] = FS.DEFAULT_POP_TYPE
        fw.pars["population type"] = FS.DEFAULT_POP_TYPE
        fw.interactions["to population type"] = FS.DEFAULT_POP_TYPE
        fw.interactions["from population type"] = FS.DEFAULT_POP_TYPE

    if proj.data:
        for pop_spec in proj.data.pops.values():
            pop_spec["type"] = FS.DEFAULT_POP_TYPE

        # Fix up TDVE types
        # Fix up transfers and interactions
        for tdve in proj.data.tdve.values():
            tdve.type = FS.DEFAULT_POP_TYPE

        for interaction in proj.data.transfers + proj.data.interpops:
            interaction.from_pop_type = FS.DEFAULT_POP_TYPE
            interaction.to_pop_type = FS.DEFAULT_POP_TYPE

    for parset in proj.parsets.values():
        parset.pop_types = [pop["type"] for pop in proj.data.pops.values()]  # If there are parsets without data, then we don't know what pop types to add. Project is essentially incomplete and considered unusable

    for result in all_results(proj):
        for pop in result.model.pops:
            pop.type = FS.DEFAULT_POP_TYPE

    return proj


@migration("Project", "1.3.0", "1.4.0", "Parameter can be derivative")
def _add_project_derivatives(proj):
    for fw in all_frameworks(proj):
        fw.pars["is derivative"] = "n"
    return proj


@migration("Result", "1.3.0", "1.4.0", "Parameter can be derivative")
def _add_result_derivatives(result):
    for pop in result.model.pops:
        for par in pop.pars:
            par.derivative = False
    return result


@migration("Project", "1.4.3", "1.5.0", "Parameters with functions can be overwritten")
def _add_parset_disable_function(proj):
    # Add skip_function flag to parset Parameter instances
    for parset in proj.parsets.values():
        for par in parset.all_pars():
            par.skip_function = sc.odict.fromkeys(par.ts, None)
    return proj


@migration("Result", "1.4.3", "1.5.0", "Parameters with functions can be overwritten")
def _result_add_skip_flag(result):
    for pop in result.model.pops:
        for par in pop.pars:
            par.skip_function = None
    return result


@migration("Project", "1.5.1", "1.5.2", "OptimInstruction has separate adjustment and start years")
def _separate_optiminstruction_years(proj):
    if hasattr(proj, "optims"):
        for optim in proj.optims.values():
            if "adjustment_year" not in optim.json:
                optim.json["adjustment_year"] = optim.json["start_year"]
    return proj


@migration("Project", "1.7.0", "1.8.0", "Parameters store interpolation method, deprecate scenario smooth onset")
def _add_parameter_interpolation_method(proj):
    for parset in proj.parsets.values():
        for par in parset.all_pars():
            par._interpolation_method = "pchip"  # New projects will default to linear, but migrations use pchip to ensure results don't change
    for scen in proj.scens.values():
        if isinstance(scen, atomica.ParameterScenario):
            for par_name in scen.scenario_values.keys():
                for pop_name in scen.scenario_values[par_name].keys():
                    if "smooth_onset" in scen.scenario_values[par_name][pop_name]:
                        logger.warning("Parameter scenario smooth onset is deprecated and will not be used")
    return proj


@migration("Project", "1.8.0", "1.9.0", "OptimInstructions functionality moved to apps")
def _refactor_optiminstructions(proj):
    if hasattr(proj, "optims"):
        delkeys = []
        for k, v in proj.optims.items():
            if isinstance(v, _Placeholder) and hasattr(v, "json"):  # If it was previously an OptimInstructions
                # This was a project from the FE, which stores the optimization JSON dicts in
                if not hasattr(proj, "optim_jsons"):
                    proj.optim_jsons = []

                json = v.json
                for prog_name in json["prog_spending"].keys():
                    spend = json["prog_spending"][prog_name]
                    try:
                        prog_label = proj.progsets[-1].programs[prog_name].label
                    except Exception:
                        prog_label = prog_name
                    json["prog_spending"][prog_name] = {"min": spend[0], "max": spend[1], "label": prog_label}
                proj.optim_jsons.append(json)
                delkeys.append(k)
        for k in delkeys:
            del proj.optims[k]
    else:
        proj.optims = atomica.NDict()  # Make sure it's defined, even if it's empty

    return proj


@migration("Project", "1.10.0", "1.11.0", "TDVE stores headings to write internally")
def _add_internal_flags_to_tdve(proj):
    # This migration adds missing attributes to TDVE and TDC objects
    if proj.data:
        for tdve in proj.data.tdve.values():
            tdve.assumption_heading = "Constant"
            tdve.write_assumption = True
            tdve.write_units = True
            tdve.write_uncertainty = True
            tdve.comment = None

        for tdc in proj.data.transfers + proj.data.interpops:
            tdc.assumption_heading = "Constant"
            tdc.write_assumption = True
            tdc.write_units = True
            tdc.write_uncertainty = True
            if not hasattr(tdc, "from_pops"):  # This was missed in the previous migration `add_pop_type` so add it in here
                # If the pop type is missing, then we must be using a legacy framework with only one pop type
                tdc.from_pop_type = list(proj.framework.pop_types.keys())[0]
                tdc.from_pops = list(proj.data.pops.keys())
                tdc.to_pop_type = list(proj.framework.pop_types.keys())[0]
                tdc.to_pops = list(proj.data.pops.keys())

    return proj


@migration("ProjectFramework", "1.12.2", "1.13.0", "Timed compartment updates to framework")
def _add_framework_timed_comps(fw):
    fw.pars["timed"] = "n"
    fw.comps["duration group"] = None
    return fw


@migration("Result", "1.12.2", "1.13.0", "Timed compartment updates to result")
def _add_result_timed_attribute(result):
    result.model.unlink()
    for pop in result.model.pops:
        for i, comp in enumerate(pop.comps):
            if comp.tag_birth:
                comp.__class__ = atomica.SourceCompartment
            elif comp.tag_dead:
                comp.__class__ = atomica.SinkCompartment
            elif comp.is_junction:
                comp.__class__ = atomica.JunctionCompartment
                comp.duration_group = None
            del comp.tag_birth
            del comp.tag_dead
            del comp.is_junction
    result.model.relink()  # Make sure all of the references are updated to the new compartment instance - it has the same ID so it should be fine
    return result


@migration("Project", "1.14.0", "1.15.0", "Refactor migration update flag")
def _rename_update_field(proj):
    if not hasattr(proj, "_result_update_required"):
        proj._result_update_required = False
    proj._update_required = proj._result_update_required
    del proj._result_update_required

    for result in all_results(proj):
        result._update_required = False
    return proj


@migration("Project", "1.15.0", "1.16.0", "Projects may change due to uncapped probabilities")
def _proj_refactor_settings_storage(proj):
    proj._update_required = True
    return proj


@migration("Result", "1.15.0", "1.16.0", "Results may change due to uncapped probabilities")
def _result_refactor_settings_storage(result):
    result._update_required = True
    return result


@migration("Project", "1.16.0", "1.17.0", "Add sim end year validation to project settings")
def _add_end_year_to_project_settings(proj):
    proj.settings = atomica.ProjectSettings(**proj.settings.__dict__)
    return proj


@migration("Project", "1.17.0", "1.18.0", "Add data TDC and TDVE attributes")
def _add_tdc_tdve_attributes(proj):
    if proj.data:
        for tdve in proj.data.tdve.values():
            tdve.ts_attributes = {}
        for tdc in proj.data.transfers + proj.data.interpops:
            tdc.attributes = {}
            tdc.ts_attributes = {}
    return proj


@migration("Project", "1.19.0", "1.20.0", "Framework.transitions is a defaultdict")
def _framework_transitions_defaultdict(proj):
    for fw in all_frameworks(proj):
        fw.transitions = defaultdict(list, fw.transitions)
    return proj


@migration("ProgramSet", "1.23.1", "1.23.2", "ProgramSet records all compartments including non-targetable ones")
def _add_progset_non_targetable_flag(progset):
    for code_name, comp in progset.comps.items():
        if not isinstance(comp, dict):
            progset.comps[code_name] = {"label": comp, "type": None}
            comp = progset.comps[code_name]
        if "non_targetable" not in comp:
            comp["non_targetable"] = False
    return progset


@migration("ProjectFramework", "1.23.4", "1.24.0", "Change framework columns to numeric types")
def _convert_framework_columns(framework):
    # This migration can be performed by simple revalidation. It will also allow
    # any other validation-related changes to be updated
    for df in framework.sheets["transitions"]:
        if not pd.isna(df.index.name):
            df.reset_index(inplace=True)

    framework._validate()
    return framework


@migration("ParameterSet", "1.28.3", "1.28.4", "Add initialization attribute")
def _parset_add_initialization(parset):
    parset.initialization = None
    return parset


@migration("ProjectData", "1.30.0", "1.31.0", "Add pop types attribute and default_all")
def _projectdata_add_types_default(D):

    if not hasattr(D, "_pop_types"):
        # We have to check for existence because
        # Are any population types present?
        pop_types = list(x["type"] for x in D.pops.values() if "type" in x)
        if not pop_types:
            pop_types = [FS.DEFAULT_POP_TYPE]

        for pop, spec in D.pops.items():
            if "type" not in spec:
                spec["type"] = pop_types[0]
        D._pop_types = list({x["type"]: None for x in D.pops.values()}.keys())  # Using a dictionary here allows for order-preserving unique

    for tdve in D.tdve.values():

        # Fix the '_add_pop_type' migration which added tdve.type instead of tdve.pop_type
        if hasattr(tdve, "type"):
            tdve.pop_type = tdve.type
            delattr(tdve, "type")
        elif not hasattr(tdve, "pop_type"):
            tdve.pop_type = D._pop_types[0]  # The majority of the time, if the pop_type is missing, the default needs to be added

        # Also add the default_all attribute
        tdve.default_all = False

        # Some old TDVE tables have a lowercase 'n.a.' instead of the correct default value
        for ts in tdve.ts.values():
            if ts.units == FS.DEFAULT_SYMBOL_INAPPLICABLE.lower():
                ts.units = FS.DEFAULT_SYMBOL_INAPPLICABLE

        # A TDVE should always have some allowed units, if the allowed units have not been populated, then draw
        # them from the ts entries. There are some older saved files that may have no allowed units even though the ts
        # entries themselves have units.
        if not hasattr(tdve, "allowed_units") or not tdve.allowed_units:
            tdve.allowed_units = list({x.units: None for x in tdve.ts.values()}.keys())

    return D



# Test timeseries utility object

import atomica as at
import numpy as np


def test_printing():
    a = at.TimeSeries([1], [5])
    print(a)


def test_equality():
    # Two time series should be equal if their values are equal
    a = at.TimeSeries()
    b = at.TimeSeries()
    assert a == b

    a.insert(None, 1)
    assert a != b

    b.insert(None, 1)
    assert a == b

    a = at.TimeSeries([1, 2, 3], [1, 2, 3])
    b = at.TimeSeries([1, 2, 3], [1, 2, 3])
    c = at.TimeSeries([1, 2, 3, 4], [1, 2, 3, 4])
    assert a == b
    assert a != c
    a.insert(None, 1)
    assert a != b


def test_constructor():
    a = at.TimeSeries()
    a.insert(1, 1)
    a.insert(2, 2)
    a.insert(None, 0)
    a.insert(3, 3)

    b = at.TimeSeries([None, 1, 2, 3], [0, 1, 2, 3])
    assert a == b

    c = at.TimeSeries(np.array([None, 1, 2, 3]), np.array([0, 1, 2, 3]))
    assert a == c

    a.insert(np.array([2, 3]), [4, 5])
    c.insert(2, 4)
    c.insert(3, 5)
    assert a == c

    d = at.TimeSeries(range(3), [0, 1, 4])
    assert a != d
    d.remove(0)
    d.insert(3, 5)
    d.insert(None, 0)
    assert a == d


def test_insert_sorting():
    a = at.TimeSeries([1, 2], [1, 2])

    # Insert at end
    a.insert(3, 3)
    assert a.t == [1, 2, 3]
    assert a.vals == [1, 2, 3]

    # Insert in middle
    a.insert(2.5, 2.5)
    assert a.t == [1, 2, 2.5, 3]
    assert a.vals == [1, 2, 2.5, 3]

    # Insert at start
    a.insert(0, 0)
    assert a.t == [0, 1, 2, 2.5, 3]
    assert a.vals == [0, 1, 2, 2.5, 3]

    # Overwrite existing at end
    a.insert(3, 4)
    assert a.t == [0, 1, 2, 2.5, 3]
    assert a.vals == [0, 1, 2, 2.5, 4]

    # Overwrite existing at start
    a.insert(0, 4)
    assert a.t == [0, 1, 2, 2.5, 3]
    assert a.vals == [4, 1, 2, 2.5, 4]

    # Overwrite existing in middle
    a.insert(1, 4)
    assert a.t == [0, 1, 2, 2.5, 3]
    assert a.vals == [4, 4, 2, 2.5, 4]


if __name__ == "__main__":

    test_constructor()
    test_printing()
    test_equality()
    test_insert_sorting()



# Test quiet context

import atomica as at

testdir = at.parent_dir()
tmpdir = testdir / "temp"


def test_file_logging():
    fname = tmpdir / "log.txt"

    at.start_logging(fname)
    at.logger.info("test")

    # Check that after logging is stopped, output is not added to the file
    at.stop_logging()
    at.logger.info("test2")
    assert open(fname).readlines()[-1].strip().split()[-1] == "test"

    at.start_logging(fname)
    at.logger.info("test3")
    at.start_logging(fname)
    at.logger.info("test4")
    at.start_logging(fname, reset=True)
    at.logger.info("test5")


def test_quiet():
    # Nothing should be printed out by this test
    logger_level = at.logger.getEffectiveLevel()

    with at.Quiet():
        at.demo("sir")

    with at.Quiet(False):
        at.demo("sir")

    assert at.logger.getEffectiveLevel() == logger_level, "Logger level has been permanently changed"


if __name__ == "__main__":
    test_quiet()
    test_file_logging()



# This script checks that parameter scale factors work correctly

import numpy as np
import atomica as at
import pytest

testdir = at.parent_dir()
tmpdir = testdir / "temp"


def test_scale_factors():
    P = at.demo("sir", do_run=False)

    res = P.run_sim("default")
    baseline = res.get_variable("infdeath")[0].vals[0]

    ps2 = P.make_parset("scaled")
    ps2.pars["infdeath"].y_factor["adults"] = 1.5
    res = P.run_sim(ps2)
    scale1 = res.get_variable("infdeath")[0].vals[0]

    ps2.pars["infdeath"].meta_y_factor = 2.0
    res = P.run_sim(ps2)
    scale2 = res.get_variable("infdeath")[0].vals[0]

    assert np.isclose(scale1, baseline * 1.5, 1e-4)
    assert np.isclose(scale2, baseline * 3, 1e-4)


def test_load_legacy_calibrations():
    # Test loading old versions of saved calibrations
    for model in ["tb", "sir", "combined"]:
        P = at.demo(model, do_run=False)
        P.parsets[0].load_calibration(testdir / "calibration" / f"test_calibration_{model}.xlsx")

    P = at.demo("combined", do_run=False)
    P.parsets[0].load_calibration(testdir / "calibration" / "test_calibration_combined_inexact.xlsx")


def test_save_load_calibrations():
    P = at.demo("tb", do_run=False)
    ps1 = P.make_parset("ps1")
    ps2 = P.make_parset("ps2")

    ps1.pars["alive"].y_factor["0-4"] = 2
    ps1.transfers["age"]["0-4"].y_factor["5-14"] = 3
    ps1.interactions["w_ctc"]["0-4"].y_factor["5-14"] = 4
    ps1.pars["alive"].meta_y_factor = 5
    ps1.transfers["age"]["0-4"].meta_y_factor = 6
    ps1.interactions["w_ctc"]["0-4"].meta_y_factor = 7

    assert ps2.pars["alive"].y_factor["0-4"] == 1
    assert ps2.transfers["age"]["0-4"].y_factor["5-14"] == 1
    assert ps2.interactions["w_ctc"]["0-4"].y_factor["5-14"] == 1
    assert ps2.pars["alive"].meta_y_factor == 1
    assert ps2.transfers["age"]["0-4"].meta_y_factor == 1
    assert ps2.interactions["w_ctc"]["0-4"].meta_y_factor == 1

    ss = ps1.calibration_spreadsheet()
    ps2.load_calibration(ss)

    assert ps2.pars["alive"].y_factor["0-4"] == 2
    assert ps2.transfers["age"]["0-4"].y_factor["5-14"] == 3
    assert ps2.interactions["w_ctc"]["0-4"].y_factor["5-14"] == 4
    assert ps2.pars["alive"].meta_y_factor == 5
    assert ps2.transfers["age"]["0-4"].meta_y_factor == 6
    assert ps2.interactions["w_ctc"]["0-4"].meta_y_factor == 7

    _ = ps2.y_factors


if __name__ == "__main__":
    # test_scale_factors()
    test_load_legacy_calibrations()
    # test_save_load_calibrations()
